/*******************************************************************************
 Module: COMredirectd.c
 
 Description: COMredirect server daemon for Linux
 
 Copyright (c) Black Box Corporation 1999-2009
 All rights reserved
 
*******************************************************************************/

/* Linux stdlib pro ttypes two functions (grantpt, unlockpt) only when
this symbol is defined. I don't know a "neater" way to do this, andma
they don't have a manual :-( -- REW */
#define _GNU_SOURCE


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
//#include <term.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <termios.h>
#include <ctype.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include <assert.h>
#include <termio.h>
#include <asm/ioctls.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>

#include <netdb.h>
#include <sys/wait.h>

#ifdef USE_SSL
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include "ssl_crd.h"
#endif /* USE_SSL */

#include "clparse.h"
#include "crd.h"
#include "pkt_forwarding.h"
#include "COMredirect.h"

//****************************************************************************
// Local constant definitions					    *
//****************************************************************************
//

#define MAX_TRACE_LEN 16
#define SMSIZEOF(_STRUCT, _MEMBER) sizeof((((_STRUCT *)0)->_MEMBER))



/****************************************************************************
 *	Global data definitions						    *
 ****************************************************************************/
forwarding_info_t	info;		// structure for packet forwarding.


#define OPMODE_NAME_OPTIMIZELAN			"optimize_lan"
#define OPMODE_NAME_LOWLATENCY			"low_latency"
#define OPMODE_NAME_PACKETIDLETIMEOUT	"packet_idle_timeout"
#define OPMODE_NAME_CUSTOM				"custom"

opmode_type_t opmode_type=OPMODE_TYPE_NONE;
/* bit definitions for DServerFeatures */
#define TPSFEAT_IOLANDS				 0x80000000		// The server is an IOLANDS that supports feature identification at session startup (ie. version >= 3.2)
#define TPSFEAT_TCPBREAKCTL		 0x40000000		// break supported on the TCP channel
#define TPSFEAT_SYNCDRAIN			 0x20000000		// synchronous drain
#define TPSFEAT_TCPXREPORT			 0x10000000		// extended report mode command (MSR & LSR info) on the TCP channel
#define TPSFEAT_NO_UDP				 0x08000000		// No UDP support
#define TPSFEAT_SYNCFLUSHENDRESET 0x04000000		// DS_TCP_SYNC_FLUSH_END reset window logic support


// COMredirect feature/logic support/configured mask
#define TPSFEAT_MASK	  (TPSFEAT_IOLANDS		| \
								TPSFEAT_TCPBREAKCTL	| \
								TPSFEAT_SYNCDRAIN 	| \
								TPSFEAT_TCPXREPORT	| \
								TPSFEAT_NO_UDP			| \
								TpFeatureSupport 		| \
								TPSFEAT_SYNCFLUSHENDRESET)	

uint32_t DServerFeatures = 0;
 
/* Command line options */
char	*fixed_ttyname  = NULL;    // Ptr to link file name

//  Setting to 1 prevents an unlink from occuring if error occures.
char	*terminal_name	= NULL;		// Ptr to terminal name
int			tcp_port 		= -1;		/* TCP port to receive on */
int			keepalive_time 	= DEFAULT_KEEPALIVE_TIME; /* Time in seconds for keepalive */
trace_level_t	trace_level     = tl_none;	// Tracing level
protocol_mode_t mode = pm_lite; 	// Operating mode


BOOL		close_tty_flag	= FALSE;	/* Hangup - close tty ports if tcp connection is lost */
int		pkt_fwd_enabled = 0;

char	*server_tcp_host = NULL;			// server mode host IP address/name
char	*client_tcp_host = NULL;			// client mode host IP address/name
int	retry_time = DEFAULT_RETRY_TIME;
int	retry_num = DEFAULT_RETRY_NUM;
int	nodisc_tcp = 0;
int	no_tcp_nagle = 0;
char  *opmode_name;				// name of operational mode ("optimize_lan","low_latency","idle_time_packet","custom")
int idle_time_packet_timeout=DEFAULT_IDLE_TIME_PACKET_TIMEOUT;
int connect_on_init=0;  // for client initiated, if set then try to establish the TCP session immediately when loaded
int open_wait_time=DEFAULT_OPEN_WAIT_TIME;
int	delay_close_tcp_time = DEFAULT_DELAY_CLOSE_TCP_TIME;		// delay (in seconds) TCP close on slave close
int restorenet=1;		// restore failed network connection.  default is yes
int use_legacy_UDP=1;	// Use legacy UDP protocol for full mode, default is yes
/******* end of command line opentions ****/

/* DEVICE_SERVER connection variables */
BOOL		dserver_connected;				/* State of DEVICE_SERVER connection */
BOOL		tty_connected = 0;					/* State of main TTY connection */
unsigned char	tp_connected = 0;	// State of COMredirect connection, which include TCP, SSL and fullmode 
state_t	state;              		// State of Device Server protocol
int		dserver_fd;					// DEVICE_SERVER TCP socket
int		ipv6_not_running = 0;
int flushing_net_read=0;			// <>0 means we're discarding any data read from the network (not commands)

struct	sockaddr_in6		dserver_addr6;    	/* Address details of Device Server (remote end) */
struct	sockaddr_in6		udp_addr6;          /* address for UDP messages */
struct	sockaddr_in		dserver_addr; 	    /* Address details of Device Server (remote end) */

int		udp_sock = -1;

struct sockaddr_in		udp_addr;           /* address for UDP messages */
unsigned char   udp_sequence_no;        /* Sequence number for UDP messages */
udp_response_t  udp_response;

int 	dserver_port_no;
int current_state = 0;
int	out_window_size;   			// amount allowed to write to Device Server
int	in_window_used;				// amount read from Device Server

const int			send_bufsize	= LM_MAX_SEND_BUF_SIZE;	/* send window for DEVICE_SERVER socket(lite mode only) */

buf_t       dserver_in_buf;           /* data read from device server */
buf_t       dserver_out_buf;

unsigned int    max_udp_retries = 20;
int 		udp_ack_timeout = 250000;	/* Timeout for UDP commands(uS) */
int 		drainpoll_period  = 500000;	/* Delay for doing drain polls(uS) */
int		keepalive_sent = 0;			// keep-alive sent flag
int slave_is_open = FALSE;

/* we use dserver_termio_modes as the master set of flags */
struct termios  dserver_termios_modes;

/* other globals */
FILE 	*trace_file;					// trace file
int		ptyxm_fd = -1;

int		ptyx_mode       = 1;			// We only support fixed tty mode.
static char    *ptydev = NULL;  /* pty name */
int		pty_mode = -1;  /* The permissions on the pty device */
int   pty_uid  = -1;
int   pty_gid  = -1;
unsigned int ptyx_ctrl_status=0;		// ptyx ctrl_status from ptyx driver
int end2end_ctrl=1;					// end to end control active? (for drain)
int dev_type=0;
int tty_write_blocked=0;
const char	COMredirect_dir[] 		= "/etc/COMredirect";

#define DEV_TYPE_UNKNOWN 0
#define DEV_TYPE_PTY     1
#define DEV_TYPE_PTMX    2

char *tcp_host = NULL;	
int	delay_close_tcp_flag = FALSE;
int netfailed=0;		// indicates if the network connection failed.  This is used to control how we bring up the network connection

static struct sigaction sigact;

// These next 2 variables are used to optimize LAN traffic.
// We will only update the Device Server if we need to...
// if we have never sent the signal state or the signal state has
// changed
int signal_state_sent=0;
unsigned int prev_signal_state=0;

int pending_drain=0;




#ifdef	USE_SSL

static	SSL_CTX		*ssl_ctx;
static	SSL     	*ssl;
BIO 		*bio_err=NULL;		//used for debug for SSL
SSL_CFG		sslcfg;
int			read_ssl, write_ssl;

static int  tp_pem_passwd_cb(char *buf, int size, int rwflag, void *password);
static RSA  *tp_tmp_rsa_cb(SSL *s, int is_export, int keylength);

#endif /* USE_SSL */


struct  clparse_st  ptab[] =    /* Cmd argument parse table */
{
/*  Flags       Value 		Variable				Token			Sub-opt
	-----       --------   	--------            	-----  			------- */
	{CLF_NAMES,   		0,     	&fixed_ttyname,			"-tty",         	NULL },
	{CLF_NAMES,   		0,     	&terminal_name,			"-term",        	NULL },
	{CLF_WVAL,    		0,     	&tcp_port,				"-port",        	NULL },
	{CLF_WVAL,    		0,     	&keepalive_time,		"-ka",          	NULL },
	{CLF_WVAL,    		0,     	&pty_mode,				"-mode",        	NULL },
	{CLF_WVAL,    		0,     	&pty_uid,				"-uid",         	NULL },
	{CLF_WVAL,    		0,     	&pty_gid,				"-gid",         	NULL },
	{CLF_WVAL,    		0,     	&trace_level,			"-trace",       	NULL },
	{CLF_WINIT,   		4,     	&trace_level,			"-debug",       	NULL },
	{CLF_WINIT,   		5,  	&trace_level,			"-data",        	NULL },
	{CLF_WOR,     		pm_lite, &mode,					"-standard",    	NULL },
	{CLF_WOR,     		pm_full, &mode,          		"-full",	NULL },
	{CLF_WOR,     		1,   	&close_tty_flag, 		"-hub",         	NULL },
	{CLF_WOR,     		1,  	&close_tty_flag, 		"-hup",         	NULL },
	#ifdef USE_SSL	
	{CLF_WOR,     		1,     	&sslcfg.enabled,		"-ssl",         	NULL },
	#endif /* USE_SSL */
	{CLF_WOR,	  		1,		&pkt_fwd_enabled,		"-pf",				NULL },
	{CLF_NAMES,	  		0,		&server_tcp_host,		"-server",			NULL },
	{CLF_NAMES,	  		0,		&client_tcp_host,		"-client",			NULL },
	{CLF_WVAL,	  		0,		&retry_time,			"-retrytime",		NULL },
	{CLF_WVAL,	  		0,		&retry_num,				"-retrynum",		NULL },
	{CLF_WOR,	  		1,		&nodisc_tcp,    		"-nodisc",			NULL },
	{CLF_WOR,	  		1,		&no_tcp_nagle,		 	"-nagleoff",		NULL },
	{CLF_NAMES,			0,		&opmode_name,			"-opmode",			NULL },
	{CLF_WVAL,			0,		&idle_time_packet_timeout,"-pktidletime",	NULL },
	{CLF_WINIT,	  		1,		&connect_on_init, 		"-initconnect",		NULL },
	{CLF_WVAL,	  		0,		&delay_close_tcp_time,"-closedelaytime",	NULL },
	{CLF_WVAL,			0,		&open_wait_time,		"-openwaittime",	NULL },
	{CLF_WINIT,			0,		&restorenet,			"-norestorenet",	NULL },
	{CLF_WINIT,			0,		&use_legacy_UDP,		"-noudp",			NULL },
	{CLF_WINIT,			1,		&use_legacy_UDP,		"-useudp",			NULL },

	{CLF_EOT,     		0,          NULL,               NULL,         		NULL }
};

/****************************************************************************
 *	Definitions used for tracing					    *
 ****************************************************************************/
char    *trace_level_str[] =
    {
        "tl_none  ",
        "tl_error ",
        "tl_status",
        "tl_info  ",
        "tl_debug ",
        "tl_data  ",
    };

/* Disable special character functions */
#ifdef _POSIX_VDISABLE
# define VDISABLE   _POSIX_VDISABLE
#else
# define VDISABLE   255
#endif


typedef struct msg_name
{
	int  msg;
	char *name;
} msg_name_t;


/* Device Server UDP messages */
msg_name_t dserver_udp_msg_list[] =
    {
        {DS_UDP_RESET, 				"DS_UDP_RESET"					},
        {DS_UDP_TCXON,                   "DS_UDP_TCXON"   },
        {DS_UDP_TCXOFF,                  "DS_UDP_TCXOFF"  },
        {DS_UDP_WFLUSH,                  "DS_UDP_WFLUSH"  },
        {DS_UDP_RFLUSH,                  "DS_UDP_RFLUSH"  },
        {DS_UDP_WR_FLUSH,                "DS_UDP_WR_FLUSH"    },
        {DS_UDP_ACK,                     "DS_UDP_ACK" },
        {DS_UDP_DRAIN_POLL,              "DS_UDP_DRAIN_POLL"  },
        {DS_UDP_DRAIN_CLEAR,             "DS_UDP_DRAIN_CLEAR" },
        {DS_UDP_KEEP_ALIVE,              "DS_UDP_KEEP_ALIVE"  },
        {DS_UDP_NAK,                     "DS_UDP_NAK" },
        {DS_UDP_SYN,                     "DS_UDP_SYN" },
        {DS_UDP_WAK,                     "DS_UDP_WAK" },
        {	DS_UDP_GET_MODEM_STATUS,"DS_UDP_GET_MODEM_STATUS"	},
        {	DS_UDP_GET_COMM_STATUS,	"DS_UDP_GET_COMM_STATUS"	},
        {	DS_UDP_LSRMST_INSERT,	"DS_UDP_LSRMST_INSERT"		},
        {	DS_UDP_SET_BREAK_ON,		"DS_UDP_SET_BREAK_ON"		},
        {	DS_UDP_SET_BREAK_OFF,	"DS_UDP_SET_BREAK_OFF"		},
        {	DS_UDP_IMMEDIATE_CHAR,	"DS_UDP_IMMEDIATE_CHAR"		},
        {	DS_UDP_SET_DTR,       	"DS_UDP_SET_DTR"				},
        {	DS_UDP_CLR_DTR,       	"DS_UDP_CLR_DTR"				},
        {	DS_UDP_SET_RTS,       	"DS_UDP_SET_RTS"				},
        {	DS_UDP_CLR_RTS,       	"DS_UDP_CLR_RTS"				},
        {	DS_UDP_GET_DTRRTS,      "DS_UDP_GET_DTRRTS"			},
        {	-1, 							"Unknown"						}
    };

/* DEVICE_SERVER TCP Messages(in-band) */
msg_name_t dserver_tcp_msg_list[] =
    {
        {DS_TCP_SEND_BREAK,      "DS_TCP_SEND_BREAK"  },
        {DS_TCP_STTY_SET,        "DS_TCP_STTY_SET/DS_TCP_CD1400_STATE" },
        {DS_TCP_STTY_REPORT_ON,  "DS_TCP_STTY_REPORT_ON"  },
        {DS_TCP_STTY_REPORT_OFF, "DS_TCP_STTY_REPORT_OFF" },
        {DS_TCP_SYNC_FLUSH_END,  "DS_TCP_SYNC_FLUSH_END"  },
        {DS_TCP_PORT_NUMBER,     "DS_TCP_PORT_NUMBER" },
        {DS_TCP_DRAIN_PORT,      "DS_TCP_DRAIN_PORT"  },
        {DS_TCP_DRAIN_ACK,       "DS_TCP_DRAIN_ACK"   },
        {DS_TCP_LSR_NODATA,      "DS_TCP_LSR_NODATA" },
        {DS_TCP_MST,             "DS_TCP_MST" },
        {DS_TCP_SEND_BREAKON,    "DS_TCP_SEND_BREAKON"},
        {DS_TCP_SEND_BREAKOFF,   "DS_TCP_SEND_BREAKOFF"},
        {	DS_TCP_SEND_SYNCDRAIN,	"DS_TCP_SEND_SYNCDRAIN"	},
        {	DS_TCP_STTY_XREPORT_ON,	"DS_TCP_STTY_XREPORT_ON"	},
        {	DS_TCP_FEATURE_ACK,		"DS_TCP_FEATURE_ACK"	},
        {	DS_TCP_SET_DTR,			"DS_TCP_SET_DTR"	},
        {	DS_TCP_CLR_DTR,			"DS_TCP_CLR_DTR"	},
        {	DS_TCP_SET_RTS,			"DS_TCP_SET_RTS"	},
        {	DS_TCP_CLR_RTS,			"DS_TCP_CLR_RTS"	},
        {DEVICE_SERVER_DATA,     "DEVICE_SERVER_DATA"    },
        {DEVICE_SERVER_NONE,     "DEVICE_SERVER_NONE"    },
        {-1,                        	"Unknown"   },
    };







/****************************************************************************
 *	Main program							    *
 ****************************************************************************/
/*
	Arguments	: - See documentation
	Return values	: - Exit code
*/

int main(int argc, char **argv)
{
	extern char 	*ptsname();
	struct timeval	timeout;
	fd_set			read_fds;	/* Arrays for select call */
	fd_set			write_fds;	/* Arrays for select call */
	fd_set			exception_fds;	/* Array for select on error */
	int				ret;
	buf_t2			buffer;
	int				len;
	int				read_flag;
	struct timeval 	*ptimeout = NULL;
	time_t			cur_time;
	int				force_ptyx_write=0;
	char 				*tmpptr, *tmp_tcp_host;
	char				tcp_host_str[MAX_HOST_NAMELEN +1] = {0};
	/* termios */
	struct termios  new_dserver_termios_modes;
	unsigned int new_state;
	
	//	struct stat		*buf;
#ifdef USE_SSL
	int         	ssl_pending = 0;
#endif	//USE_SSL

	// Must be root - this is somewhat redundant as the crdadm will normally
	//	be used to start the daemon and it has this check as well. However I
	//	will keep it here for completness.
	if (getuid() != 0)
	{
		fprintf( stderr, "Not root !\n" );
		exit(FAILURE);
	}

	ret = clparse(argc, argv, ptab);

	if (trace_level > tl_none)
	{
		char file_name[80];
//		sprintf( file_name, "%s/trace.%d", COMredirect_dir, tcp_port );
		// strip off dev
		char *master_tty_name = strrchr(fixed_ttyname, '/'); 
		sprintf(file_name, "%s/trace.%s", COMredirect_dir, master_tty_name + 1 );

		trace_file = fopen(file_name, "a");
		if (trace_file == NULL)
		{
			perror("COMredirect error: opening trace file");
			exit(EXIT_FAILURE);
		}
		cur_time = time( &cur_time );
		trace( tl_info, "\n******************:    %s", ctime( &cur_time ));
	}

	if ( ret	!= 0 )
	{
		trace(tl_error, "invalid option: %s\n", argv[ret]);
		exit(EXIT_FAILURE);
	}

	if (tcp_port == -1)
	{
		trace(tl_error, "TCP port not specified\n");
		exit(-1);
	}

	trace(tl_status, "COMredirect starting Version %s on port %d", TP_VERSION, tcp_port);
	trace(tl_status, " - (c) 1999-2009, Black Box Corporation\n" );

	if (opmode_name)
	{
		if (!strcmp(opmode_name, OPMODE_NAME_OPTIMIZELAN))
		{
			opmode_type = OPMODE_TYPE_OPTIMIZELAN;
			pkt_fwd_enabled=0;
			no_tcp_nagle=0;
		}
		else if (!strcmp(opmode_name, OPMODE_NAME_LOWLATENCY))
		{
			opmode_type = OPMODE_TYPE_LOWLATENCY;
			pkt_fwd_enabled=0;
            no_tcp_nagle=1;
		}
		else if (!strcmp(opmode_name, OPMODE_NAME_PACKETIDLETIMEOUT))
		{
			opmode_type = OPMODE_TYPE_PACKETIDLETIMEOUT;
			pkt_fwd_enabled=0;
			no_tcp_nagle=1;
		}
		else if (!strcmp(opmode_name, OPMODE_NAME_CUSTOM))
		{
			opmode_type = OPMODE_TYPE_CUSTOM;
		}
		else
		{
		   trace(tl_error, "Invalid operational mode: %s\n",opmode_name);
		   exit(-1);
		}
	}

	tcp_host = server_tcp_host;
	// if client initiated then init some variables
	if (client_tcp_host)
	{
		tcp_host = client_tcp_host;
		trace(tl_status, " - running Client Initiated mode \n");
		trace(tl_status, "   - number of reconnect retries are %d", retry_num);
		trace(tl_status, "   - reconnect interval of %d seconds \n",retry_time);
	}
	else
	{
		trace(tl_status, " - running Server mode \n");
	}

	if (tcp_host != NULL)
	{
		// copy configured IP address
		strncpy(&tcp_host_str[0], tcp_host, MAX_HOST_NAMELEN);
 		tmp_tcp_host = &tcp_host_str[0];

		// strip of IPv6 brackets if required
		tmpptr = strchr(tmp_tcp_host, '[');
		// we have a literal IPv6 address
		if (tmpptr != NULL)
		{
			*tmpptr = 0;						// null out startin "[" 
			tmp_tcp_host = tmpptr + 1;			// increment pass "["
	 
			// now search for ending "]"
			tmpptr = strchr(tmp_tcp_host, ']');
			if (tmpptr == NULL)
			{
				trace (tl_error,"Specified IPv6 address must be inclosed in square brackets]\n");
				exit(-1);
			}
			*tmpptr = 0;		// zero out ending ] bracket
			// set client or server host back
			if (client_tcp_host)
				client_tcp_host = tmp_tcp_host;
			if (server_tcp_host)
				server_tcp_host = tmp_tcp_host;
		}
	}
	
	{
		trace(tl_status, " - running %s mode\n", mode == pm_lite ? "lite " : "full" );
	}

	if (mode == pm_full)
	{
		if (use_legacy_UDP)
		{
			trace(tl_status, " - Use legacy UDP protcol for Full mode \n");
		}
		else
		{
			trace(tl_status, " - Don't use UDP procotol if device server supports it \n");
		}
}

#ifdef	USE_SSL

	trace(tl_status, " - running %s(%x)\n", OPENSSL_VERSION_TEXT, SSLeay() );

	if (bio_err == NULL)
		bio_err = BIO_new_fp( trace_file, BIO_NOCLOSE );
#endif	//USE_SSL

	/* Parameters OK, initialise ttys */
	ptyxm_fd = -1;
	tty_connected = FALSE;
	buffer.len = 0;

	/* Install signal handler */
	sigact.sa_handler = handle_signal;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = 0;
	sigaction(SIGTERM,&sigact,0);
	sigaction(SIGINT,&sigact,0);
	signal(SIGPIPE, SIG_IGN);

	dserver_connected = FALSE;
	buf_init(&dserver_out_buf);
	buf_init(&dserver_in_buf);
	
	while( 1 )
	{
		if (!tty_connected)
		{
			/* Open main tty */
			ptyxm_fd = open_tty();
			if (ptyxm_fd < 0)
			{
				trace(tl_error, "Error opening master  device \n");
                sleep(1);   // try again later
                continue;
			}
			else
			{
				tty_connected = TRUE;
	 			// inform driver of open wait time configuration and COMredirect connection state
				write_tty_ctrl(ptyxm_fd, TP_PROTO_SET_OPEN_WAIT, (unsigned char *)&open_wait_time, sizeof(int));
				is_tp_conn_avail();		// update tp_connected
				write_tty_ctrl(ptyxm_fd, TP_PROTO_TP_CONNECTION, &tp_connected, 1);
			}
		}
		if ( !dserver_connected )
		{
			initialise_connection_state ();
			dserver_fd = wait_for_dserver_connection();

			if(dserver_fd == FAILURE)
			{
				trace(tl_debug, "Connecting or Accepting dserver connection failed \n");
				sleep(3);		// just so log files won't fill up to quickly
				continue;
			}
			if (mode == pm_full)
			{
				char ch;

				// Delay a bit before sending packet mode request to make sure
				// other side has finished initializing socket or SSL
				usleep( 100 * 1000);	

				trace(tl_info, "send packet mode req\n");
				ch = DS_URG_TCP_PACKETMODE_REQ;

				// Do not send this in SSL stream if configured. The OOB data
				//  must NOT be encrypted or put in line with other data or the
				//  device server will not be able to detect it. There is such a
				//  a small amount of OOB data that encryption is not justified.
				ret = send (dserver_fd, &ch, 1, MSG_OOB);

				if (ret != 1)
				{
					trace(tl_error, "Error sending packet\n");
				}
				fflush(stdout);
				state = s_sent_packet_req;
			}
			else
			{
				state = s_data_transfer;
			}
		}

		while (dserver_connected && tty_connected)
		{
			/* Initialise FD sets for select */
			FD_ZERO(&read_fds);
			FD_ZERO(&write_fds);
			FD_ZERO(&exception_fds);

			// Test SSL enabled and pending flag. If there is SSL data pending then
			//	do not call select, fall through to the read section to get SSL
			//	pending data.
			read_flag = FALSE;
#ifdef  USE_SSL

			if( sslcfg.enabled )
			{
				if ( (ssl_pending = SSL_pending( ssl )) )
				{
					read_flag = TRUE;						// data is pending, skip select.
				}
			}
#endif  //USE_SSL
			// If there is no read data pending, then prepare file descriptors for select
			//	Else, skip down and read pending data.
			if( read_flag == FALSE )
			{
				if (tty_write_blocked)
				{
					FD_SET(ptyxm_fd,&write_fds);
					// don't want to read any network data
					FD_CLR(dserver_fd, &read_fds);		
				}
				else
				{
					FD_SET(dserver_fd, &read_fds);
				}

				if (udp_sock != -1)
					FD_SET(udp_sock, &read_fds);

				if (dserver_out_buf.len != 0 || buffer.len > 0 )
				{
					FD_SET(dserver_fd, &write_fds);
				}
				else 
				{
					if ((dserver_out_available(&info) > 2) && state == s_data_transfer)
					{
						FD_SET(ptyxm_fd, &read_fds);
					}
				}

				/* exceptions(TCP urgent data or ctrl messages) */
				if (udp_sock != -1)
					FD_SET(udp_sock, &exception_fds);

				FD_SET(dserver_fd, &exception_fds);
				FD_SET(ptyxm_fd, &exception_fds);


				// set_time_out() will return address of timeval structure if
				// timers are to be set. Otherwise, a NULL pointer is returned
				//	for no timeout.
				ptimeout = set_time_out( &info, &timeout );
				ret = select( FD_SETSIZE, &read_fds, &write_fds, &exception_fds, ptimeout );

				// timeout occured
				if (ret == 0)
				{
					// check keep-alive timers
					if ( keepalive_time )
					{
						// restart keep-alive timer if expired.
						if (check_time_out( &info.keepalive_start, keepalive_time * 1000)) 
						{
							// if not using UDP then send keep-alive immediate command
							// only do keep alive logic if in data transfer state
							if ( IsNoUDP() && (state == s_data_transfer) )
							{
								// if we've already sent one without a response then close connection
								// we will ignore keep-alive expiring if we are blocked on tty writes
								if (keepalive_sent && !tty_write_blocked)
								{
									trace(tl_info, "No keep-alive response, Closing connection\n");
									close_connection();
									keepalive_sent = 0;
								}
								else
								{
									ret = write_dserver_imm_ctrl(dserver_fd, DS_UDP_KEEP_ALIVE, 0);
									if (ret < 0)
									{
										trace(tl_error, "Error sending Keep-alive immediate cmd, closing connection\n");
										close_connection();
									}
									else
									{
										 keepalive_sent = 1;
									}
								}
							}
							// otherwise send UDP keep-alive
							// only do keep alive logic if in data transfer satae
							else if (state == s_data_transfer)
							{
								ret = write_dserver_udp_ctrl_wait(DS_UDP_KEEP_ALIVE);
								if (ret == -1 || udp_response.command == DS_UDP_SYN)
								{
								  trace(tl_error, "No UDP keep-alive response, Closing connection\n");
								  close_connection();
								}
							}
							gettimeofday( &info.keepalive_start, NULL );
						} // end of keep-alive expired

					} // end of keep-alive logic
					// check pkt forwarding timers
					if ( info.forwarding & PKT_FORWARD_ON_IDLE && info.data_count != 0 )
					{
						if ( check_time_out( &info.idle_time_start, info.idle_time ) )
						{
							trace(tl_data, "Forwarding frame on PKT_FORWARD_ON_IDLE\n");
							forward_frame(&info );
						}
					}

					if ( info.forwarding & PKT_FORWARD_ON_TIME && info.data_count != 0 )
					{
						if ( check_time_out( &info.force_time_start, info.force_time ) )
						{
 							trace(tl_data, "Forwarding frame on PKT_FORWARD_ON_TIME\n");
							forward_frame(&info );
						}
					}
					// check client-initiated delay closed timer
					if (client_tcp_host)
					{
						 if ( check_time_out( &info.delay_close_tcp_start, delay_close_tcp_time * 1000 ) )
						 {
							  if ( delay_close_tcp_flag && !(dserver_out_buf.len != 0 || buffer.len > 0 ) )
							  {
									delay_close_tcp_flag = FALSE;
									trace(tl_info, "Delayed closing of TCP connection\n");
									close_connection();
									netfailed=0;
							  }
						 }
					}
					continue;
				} // end of if time-out occured
				else if (ret < 0 )		// select error
				{
					trace(tl_error, "Error in select: %d\n", ret);
					close_connection();
					continue;
				}
			}
			// if we've lost either network or TTY then exit while connected loop now
			if (dserver_connected == 0 || tty_connected == 0)
			{
				continue;
			}
	
		
			//****************************************************************
			// Read from application through master tty (and write to network)
			//****************************************************************
			if(FD_ISSET(ptyxm_fd, &exception_fds))
			{
				/* Error but make tl_debug because no action taken */
				trace(tl_debug, "ptyxm_fd exception\n");
				usleep(100000);
			}

			if ( FD_ISSET(ptyxm_fd, &read_fds) )
			{
				len = dserver_out_available(&info)-2+1;
				buffer.data_ptr = &buffer.data[0];
				buffer.len = read( ptyxm_fd, buffer.data_ptr, len );		// read from master TTY

				/* If we get a read error, the most likely reason is that other
				   side has closed /dev/tx? . In any case, close the master
				   side and let the main loop re-open it to prevent continually
				   getting the error.
				*/
				if (buffer.len == -1)
				{
					int closeit=0;

					if (errno == EIO)
					{
						trace(tl_info, "Port closed from other side - close our side");
						closeit=1;
					}
					else
					{
						if (errno != EAGAIN)
						{
							trace(tl_error, "Error reading from master tty. errno: %d", errno);
							closeit=1;
						}
					}
					if (closeit)
					{
						 close_tty();
					}
					continue;
				}

				if (!buffer.len)
				{
					trace(tl_data, "EOF in master tty\n");
					exit(-1);
				}
				if(trace_level >= tl_data)
				{
					 trace(tl_data, "main(): read from master tty returned %d bytes", buffer.len );
					 dump_buf(trace_file, buffer.data_ptr, buffer.len );
				}

				if (buffer.data[0] == TIOCPKT_DATA)
				{
					buffer.data_ptr++;
					buffer.len--;
					process_packet_forwarding( &info, &buffer);
				}
				else
				{
					if (buffer.data[0] & TIOCPKT_FLUSHREAD)
					{
						if (mode == pm_full)
						{
							/*
							- put the daemon in read flush mode.  When in this mode
							  any received data is discarded
							- send a UDP DS_UDP_RFLUSH command to the server
							- when the daemon receives a TCP DS_TCP_SYNC_FLUSH_END
							  command from the server, exit read flush mode.
							*/
							trace(tl_debug, "FLUSHREAD from app\n");
							flushing_net_read=1;
							ret = write_dserver_udp_ctrl_wait(DS_UDP_RFLUSH);
							if (ret < 0)
							{
								trace(tl_error, "TIOCPKT_FLUSHREAD: Error sending DS_UDP_RFLUSH, closing connection\n");
								close_connection();
							}
						}
					}
					if (buffer.data[0] & TIOCPKT_FLUSHWRITE)
					{
						if (mode == pm_full)
						{
							/*
							- send a UDP DS_UDP_WFLUSH command to the server
							- send a TCP DS_TCP_SYNC_FLUSH_END command/marker to the
							  server.
							*/
							trace(tl_debug, "FLUSHWRITE from app\n");
							ret = write_dserver_udp_ctrl_wait(DS_UDP_WFLUSH);
							forward_frame(&info );	// push any queued up data
							if (ret >= 0)
								 ret = write_dserver_ctrl(dserver_fd, DS_TCP_SYNC_FLUSH_END);
							if (ret < 0)
							{
								 trace(tl_error, "TIOCPKT_FLUSHWRITE: Error sending DS_TCP_SYNC_FLUSH_END, closing connection\n");
								 close_connection();
							}
						}
					}
					if (buffer.data[0] & TIOCPKT_XCTRLSTATUS)
					{
						typeof(ptyx_ctrl_status) temp_status;
						ret=ioctl(ptyxm_fd,PTX_IOCXCSGET,&temp_status);	// get extended control status
						if (ret)
						{
							trace(tl_error, "Error in ioctl-PTX_IOCXCSGET: %d %d %d\n", __LINE__,ret, errno);
							exit(-1);
						}
						ptyx_ctrl_status |= temp_status;
						trace(tl_debug, "PTX_IOCXCSGET: %x %x\n", temp_status,ptyx_ctrl_status);
					}
					// reset buffer.len if not data 
					buffer.len = 0;


					if (ptyx_ctrl_status & XCTRLSTATUS_STOP)
						{
							ptyx_ctrl_status &= (~XCTRLSTATUS_STOP);
							trace(tl_info, "XCTRLSTATUS_STOP recieved\n");
					 
							if (mode == pm_full)
							{	
								ret = write_dserver_udp_ctrl_wait( DS_UDP_TCXOFF );
								if (ret < 0)
								{
									trace(tl_error, "XCTRLSTATUS_STOP: Error sending DS_UDP_RFLUSH, closing connection\n");
									close_connection();
								}
							}
						}
						if (ptyx_ctrl_status & XCTRLSTATUS_START)
						{
							ptyx_ctrl_status &= (~XCTRLSTATUS_START);
							trace(tl_info, "XCTRLSTATUS_START recieved\n");
					 
							if (mode == pm_full)
							{	
								ret = write_dserver_udp_ctrl_wait( DS_UDP_TCXON );
								if (ret < 0)
								{
									trace(tl_error, "XCTRLSTATUS_START: Error sending DS_UDP_RFLUSH, closing connection\n");
									close_connection();
								}
							}
						}
						if (ptyx_ctrl_status & XCTRLSTATUS_FLUSHREAD)
						{
							ptyx_ctrl_status &= (~XCTRLSTATUS_FLUSHREAD);
							trace(tl_info, "XCTRLSTATUS_FLUSHREAD recieved\n");
					 
							if (mode == pm_full)
							{
								/*
									put the daemon in read flush mode.  When in this mode
									any received data is discarded
									send a UDP DS_UDP_RFLUSH command to the server
									- when the daemon receives a TCP DS_TCP_SYNC_FLUSH_END
									command from the server, exit read flush mode.
								*/
								trace(tl_debug, "FLUSHREAD from app\n");
								flushing_net_read=1;
								ret = write_dserver_udp_ctrl_wait(DS_UDP_RFLUSH);
								if (ret < 0)
								{
									trace(tl_error, "XCTRLSTATUS_FLUSHREAD: Error sending DS_UDP_RFLUSH, closing connection\n");
									close_connection();
								}
							}
						}
						if (ptyx_ctrl_status & XCTRLSTATUS_FLUSHWRITE)
						{
							ptyx_ctrl_status &= (~XCTRLSTATUS_FLUSHWRITE);
							trace(tl_info, "XCTRLSTATUS_FLUSHWRITE recieved\n");
					 
							if (mode == pm_full)
							{
								/*
								- send a UDP DS_UDP_WFLUSH command to the server
								- send a TCP DS_TCP_SYNC_FLUSH_END command/marker to the
									server.
								*/
								trace(tl_debug, "FLUSHWRITE from app\n");
								ret = write_dserver_udp_ctrl_wait(DS_UDP_WFLUSH);
								forward_frame(&info );	// push any queued up data
								if (ret >= 0)
									ret = write_dserver_ctrl(dserver_fd, DS_TCP_SYNC_FLUSH_END);
								if (ret < 0)
								{
									trace(tl_error, "XCTRLSTATUS_FLUSHWRITE: Error sending DS_TCP_SYNC_FLUSH_END, closing connection\n");
									close_connection();
								}
							}
						}
					if (ptyx_ctrl_status & XCTRLSTATUS_CLOSE)
					{
						ptyx_ctrl_status &= (~XCTRLSTATUS_CLOSE);
						trace(tl_info, "XCTRLSTATUS_CLOSE recieved\n");
						forward_frame(&info );	// push any queued up data
						slave_is_open = FALSE;
						if (client_tcp_host)
						{
							// setup to close TCP connection on slave tty being closed
							//  as long as -nodisc not set
							if ( !nodisc_tcp && dserver_connected)
							{
								delay_close_tcp_flag = TRUE;
								gettimeofday( &info.delay_close_tcp_start, NULL );
							 }
						}
					}
					if (ptyx_ctrl_status & XCTRLSTATUS_OPEN)
					{
 						trace(tl_debug, "XCTRLSTATUS_OPEN recieved\n");
						ptyx_ctrl_status &= (~XCTRLSTATUS_OPEN);
						slave_is_open = TRUE;
						if (client_tcp_host)
						{
							 // clear delay_close_tcp_flag since we will eventually
							 // get another close after this open
							 delay_close_tcp_flag = FALSE;
						}
					}
				}
			}
			// make sure we are in data transfer state if full mode or 
			// we can't process some of these
    		if (mode == pm_lite  || ( state == s_data_transfer) )    
			{
				if (ptyx_ctrl_status & XCTRLSTATUS_TERMIOS)
				{
					trace(tl_debug, "XCTRLSTATUS_TERMIOS recieved\n");
					ptyx_ctrl_status &= (~XCTRLSTATUS_TERMIOS);
					if (mode == pm_full)
					{

						 tcgetattr(ptyxm_fd, &new_dserver_termios_modes);
						 if (memcmp(&new_dserver_termios_modes, &dserver_termios_modes, sizeof (struct termios)))
						 {
							 trace(tl_debug, "setting termios\n");
							 memcpy(&dserver_termios_modes, &new_dserver_termios_modes, sizeof (struct termios));
							 if (UseNewPktFwdLogic(&info))
							 {
								  // indicate that set tty command should be put into pkt fwd buf
								  info.dscmd = DS_TCP_STTY_SET;
							 }
							 set_dserver_modes(1);
						 }
					}
				}
				if (ptyx_ctrl_status & XCTRLSTATUS_SIGNALS)
				{
 					trace(tl_debug, "XCTRLSTATUS_SIGNALS recieved\n");
					ptyx_ctrl_status &= (~XCTRLSTATUS_SIGNALS);
					if (mode == pm_full)
					{

						 /* Siganals */
						 new_state = 0;
						 ret = -1;
						 ret = ioctl(ptyxm_fd, TIOCMGET, &new_state);	// get modem status
						 if (ret)
						 {
							 trace(tl_error, "Error in ioctl-5: %d %d\n", ret, errno);
						 }
						 else
						 {
							 if (!signal_state_sent || 
								  ((prev_signal_state & (TIOCM_RTS|TIOCM_DTR))!=(new_state & (TIOCM_RTS|TIOCM_DTR))))
							 {
								 signal_state_sent=1;
								 prev_signal_state=new_state;
								 trace( tl_debug, "XCTRLSTATUS_SIGNALS: new_state=%08x\n", new_state );
								 if (UseNewPktFwdLogic(&info))
								 {
									  // indicate that signals command should be put into pkt fwd buf
									  // don't know which one here
									  info.dscmd = DS_TCP_SET_DTR + DS_TCP_CLR_DTR + DS_TCP_SET_RTS + DS_TCP_CLR_RTS;
								 }
								 set_RTS_DTR(new_state & TIOCM_RTS, new_state & TIOCM_DTR);
								 if ( dserver_connected && ((DServerFeatures & TPSFEAT_IOLANDS) == 0) )
								 {
									 write_dserver_ctrl(dserver_fd, DS_TCP_STTY_REPORT_ON);
								 }
							 }
						 }
					}
				}
				if (ptyx_ctrl_status & XCTRLSTATUS_BREAK )
				{
					trace(tl_debug, "XCTRLSTATUS_BREAK recieved\n");
					ptyx_ctrl_status &= (~XCTRLSTATUS_BREAK);
					forward_frame(&info );	// push any queued up data
 					if (mode == pm_full)
					{
						ret = write_dserver_ctrl(dserver_fd, DS_TCP_SEND_BREAK);
						if (ret < 0)
						{
							trace(tl_error, "XCTRLSTATUS_BREAK: Error sending DS_TCP_SEND_BREAK, closing connection\n");
							close_connection();
						}
					}
				}

				if (ptyx_ctrl_status & XCTRLSTATUS_BREAKON)
				{
					trace(tl_debug, "XCTRLSTATUS_BREAKON recieved\n");
					ptyx_ctrl_status &= (~XCTRLSTATUS_BREAKON);
					forward_frame(&info );	// push any queued up data
					if (mode == pm_full)
					{
						if (DServerFeatures & TPSFEAT_TCPBREAKCTL)
						{
							ret = write_dserver_ctrl(dserver_fd, DS_TCP_SEND_BREAKON);
						}
						else
						{
							ret = write_dserver_udp_ctrl_wait(DS_UDP_SET_BREAK_ON);
						}
						if (ret < 0)
						{
							trace(tl_error, "XCTRLSTATUS_BREAKON: Error sending commands, closing connection\n");
							close_connection();
						}
					}
				}
				if (ptyx_ctrl_status & XCTRLSTATUS_BREAKOFF)
				{
 					trace(tl_debug, "XCTRLSTATUS_BREAKOFF recieved\n");
					ptyx_ctrl_status &= (~XCTRLSTATUS_BREAKOFF);
					forward_frame(&info );	// push any queued up data
					if (mode == pm_full)
					{
						if (DServerFeatures & TPSFEAT_TCPBREAKCTL)
						{
							ret = write_dserver_ctrl(dserver_fd, DS_TCP_SEND_BREAKOFF);
						}
						else
						{
							ret = write_dserver_udp_ctrl_wait(DS_UDP_SET_BREAK_OFF);
						}
						if (ret < 0)
						{
							trace(tl_error, "XCTRLSTATUS_BREAKOFF: Error sending commands, closing connection\n");
							close_connection();
						}
					}
				}
				if (ptyx_ctrl_status & XCTRLSTATUS_DRAIN)
				{
  					trace(tl_debug, "XCTRLSTATUS_DRAIN recieved\n");
					ptyx_ctrl_status &= (~XCTRLSTATUS_DRAIN);
					forward_frame(&info );	// push any queued up data
					if (DServerFeatures & TPSFEAT_SYNCDRAIN)			// note lite mode won't have this set
					{
			 			if (end2end_ctrl && tty_write_blocked==0)
						{
							ret = write_dserver_ctrl(dserver_fd, DS_TCP_SEND_SYNCDRAIN);
							pending_drain=1;
							if (ret < 0)
							{
								trace(tl_error, "XCTRLSTATUS_DRAIN: Error sending DS_TCP_SEND_SYNCDRAIN, closing connection\n");
								close_connection();
							}	
						}
						else	// local control; so fake out an ack to the pty driver
						{
//							ret = ioctl(ptyxm_fd, PTX_IOCDRAINEND, 0);	// drain complete
				 			ret = write_tty_ctrl(ptyxm_fd, TP_PROTO_DRAIN_ACK, NULL, 0); // drain complete
				 			if (ret)
				 			{
								trace(tl_error, "Error in TP_PROTO_DRAIN_ACK: %d %d %d\n", __LINE__,ret, errno);
					 			exit(-1);
				 			}
			 			}
		 			}
		 			else	// lite mode or syncdrain not supported by server
		 			{		// fake ack to the pty driver
//						ret = ioctl(ptyxm_fd, PTX_IOCDRAINEND, 0);	// drain complete
			 			ret = write_tty_ctrl(ptyxm_fd, TP_PROTO_DRAIN_ACK, NULL, 0); // drain complete
			 			if (ret)
			 			{
				  			trace(tl_error, "Error in TP_PROTO_DRAIN_ACK: %d %d %d\n", __LINE__,ret, errno);
				  			exit(-1);
			 			}
		 			}
	 			}
 			}
			
			//****************************************************************
			// Write to device server through network
			//	(only used if TCP resources not available after tty read)
			//****************************************************************
			if ( (dserver_fd != -1) && FD_ISSET(dserver_fd, &write_fds))
			{
				if (dserver_out_buf.len > 0)
				{

					ret = write_buf_raw(  dserver_fd,
					                        &dserver_out_buf.data[dserver_out_buf.offset],
					                        dserver_out_buf.len);
					if(ret >= 0)
					{
						dserver_out_buf.len -= ret;
						dserver_out_buf.offset += ret;
					}
					else
					{
						close_connection();
					}
				}
				// If dserver_out_buf has been fully sent, then check if there
				//	is any other data that needs to be forwarded.
				if ( dserver_out_buf.len == 0 && buffer.len > 0 )
				{
					process_packet_forwarding( &info, &buffer);
				}
			}

			//***********************************************************
			//	Exception from Device Server
			//****************************************************************
			if ( (dserver_fd != -1) && (FD_ISSET(dserver_fd, &exception_fds)) )
			{
				trace(tl_error, "device server exception, %d: %s\n", errno, strerror(errno));
				close_connection();
				continue;
			}

			if(udp_sock != -1)
			{
				if (FD_ISSET(udp_sock, &exception_fds))
				{
					trace(tl_debug, "UDP exception\n");
				}
				
				if (FD_ISSET(udp_sock, &read_fds))
				{
					/* we don't really expect to get any unsolicited responses */
					ret = read_dserver_udp_message(&udp_response);
					if (ret == FAILURE || udp_response.command == DS_UDP_SYN)
					{
						close_connection();
						continue;
					}
				}
			}

			//****************************************************************
			//  Read from device server through network (and write to tty)
			//****************************************************************
			read_flag = FALSE;
#ifdef USE_SSL
			if ( sslcfg.enabled )
			{
				if ( ssl_pending )
					read_flag = TRUE;		// SSL data is pending so read it.
				else if ((dserver_fd != -1) && FD_ISSET(dserver_fd, &read_fds) )
					read_flag = TRUE;
			}
			else
#endif  //USE_SSL
			{
				if ((dserver_fd != -1) && FD_ISSET(dserver_fd, &read_fds))
					read_flag = TRUE;
			}

			if ( read_flag )
			{
				int bytes_read;
				dserver_packet_t  packet_type=DEVICE_SERVER_NONE;

				bytes_read = read_dserver(dserver_fd, &packet_type, &dserver_in_buf);
		
				if(bytes_read <= 0)
				{
					/* dserver socket close or socket error */
					if(bytes_read < 0)
					{
						if(errno == EAGAIN)
						{
							trace(tl_debug, "read_device_server returned %d: %s\n", errno, strerror(errno));
							continue;
						}
						else
							trace(tl_error, "read_device_server returned %d: %s\n", errno, strerror(errno));
					}
					else
					{
						trace(tl_info, "read_device_server returned %d: %s\n", errno, strerror(errno));
					}
					close_connection();
					continue;
				}
				else if (packet_type == DEVICE_SERVER_DATA)
				{
					if (flushing_net_read || slave_is_open==FALSE)
					{
						buf_init(&dserver_in_buf);
					}
					else
					{
						force_ptyx_write=1;
					}
				}
				else
				{
					buf_init(&dserver_in_buf);
					tty_write_blocked=0;
				}
				// This was added to detect a tty close caused by loss of CD from the Device Server
				if ( !tty_connected )
				{
					ptyxm_fd = -1;
					continue;
				}
			}
			
			// try to write data to the tty
			if (ptyxm_fd != -1 && (force_ptyx_write || FD_ISSET(ptyxm_fd, &write_fds)))
			{
				force_ptyx_write=0;
				if (flushing_net_read || slave_is_open==FALSE)
				{
					trace(tl_data, "TTY is closing, or doing net flush, discarding network data\n");
					buf_init(&dserver_in_buf);
				}
				else
				{
					ret=write(ptyxm_fd, &dserver_in_buf.data[dserver_in_buf.offset], dserver_in_buf.len);
					if (ret < 0)
					{
						if (errno != EAGAIN)
						{
							trace(tl_error, "Error writing to ptyx: %d\n", errno);
							buf_init(&dserver_in_buf);
							close_connection();
							break ;
						}
						else
						{
							tty_write_blocked = 1;
							if (pending_drain)
							{
								write_tty_ctrl(ptyxm_fd, TP_PROTO_DRAIN_ACK, NULL, 0);
								pending_drain = 0;
							}
						}
					}
					else
					{
						trace(tl_debug, "%s ptyx write len=%d written=%d remaining=%d\n", (ret < dserver_in_buf.len)?"Incomplete":"Completed",dserver_in_buf.len,ret,dserver_in_buf.len-ret);
						dserver_in_buf.len-=ret;
						dserver_in_buf.offset+=ret;
						if (dserver_in_buf.len == 0)
						{
							buf_init(&dserver_in_buf);
							tty_write_blocked=0;
						}
						else
						{
							tty_write_blocked=1;
							if (pending_drain)
							{
								write_tty_ctrl(ptyxm_fd, TP_PROTO_DRAIN_ACK, NULL, 0);
								pending_drain=0;
							}
						}
					}
				}
			}
		}	// while( dserver_connect && tty_connect )
	}	// while( 1 )
	close_tty();
	return 0;
}

/****************************************************************************
 *	Initialise all state information used for DEVICE_SERVER connection	    *
 ****************************************************************************/
/*
	Arguments	: none
	Return values	: none
*/
void initialise_connection_state(void)
{
	state = s_initial;
	dserver_connected = FALSE;
	tp_connected = FALSE;
	dserver_fd = -1;
	udp_sock = -1;
	dserver_port_no = 0xff;
	udp_sequence_no = 0;
	out_window_size = DEFAULT_OUT_WINDOW_SIZE;
	in_window_used = 0;
	buf_init(&dserver_in_buf);

}

//******************************************************************************
//*	Initialise a buffer   						    *
//****************************************************************************/
//
//	Arguments	: buf : a buffer
//	Return values	: None
//
void    buf_init(buf_t *buf)
{
	buf->offset = 0;
	buf->len = 0;
}


int buf_append(buf_t *buf, char *data, int len)
{
	if ((buf->offset+buf->len+len > (int)sizeof (buf->data)) || (len<0) )
	{
		trace(tl_error, "buf_append: invalid length: %d", len);
		return FAILURE;
	}

	memcpy(&buf->data [buf->offset+buf->len], data, len);
	buf->len += len;

	return OK;
}




int open_tty()
{
	int ptyxm_fd = -1;
	int len;
	int ret;
	extern char *ptsname( );
	char *fixed_dir = NULL;
	char *p;
	static int succeeded = 0;
	DIR *dirp;
	struct dirent *dp;
	char buf[1024];
	struct stat stat_buf;
	static char linkname[1024];
	struct	termios  modes;
	int 	i, tries=0;
	char	f_name[256];
	char	*slave_ttyname;
	int		minor_number;

#define MAX_TRIES 100

	/* First try to open 'our' pty device */
	if ( fixed_ttyname )
	{
		if( lstat( fixed_ttyname, &stat_buf ) < 0 ) {
			// if the device nodes do not exist, then create both the 
			//	master (fixed_ttyname) and slave (slave_ttyname).
			slave_ttyname = strdup( fixed_ttyname );
			if( (p = strstr( slave_ttyname,"px" )) == NULL ) {
				trace( tl_error, "TTY name: %s has incorrect format.\n", slave_ttyname  );
				exit(-1);
			}
			strncpy( p, "tx", 2 );
			sscanf(  p+2, "%d", &minor_number );
			trace( tl_status, "Creating device nodes: %s, %s\n", fixed_ttyname, slave_ttyname );
			sprintf( buf, "mknod %s c %d %d", fixed_ttyname, PTYX_MASTER_MAJOR, minor_number );
			if( system( buf ) == -1 ) {
				trace( tl_error, "Could not create master device node: %s\n", fixed_ttyname  );
				trace( tl_error, "  - %s\n", strerror( errno ) );
				exit(-1);
			}
			sprintf( buf, "mknod %s c %d %d", slave_ttyname, PTYX_SLAVE_MAJOR, minor_number );
			if( system( buf ) == -1 ) {
				trace( tl_error, "Could not create slave device node: %s\n", slave_ttyname  );
				trace( tl_error, "  - %s\n", strerror( errno ) );
				exit(-1);
			}
			free( slave_ttyname );
		}
		ptyxm_fd = open( fixed_ttyname, O_RDWR | O_NONBLOCK);
		while( succeeded && (ptyxm_fd < 0) && (++tries < MAX_TRIES) )
		{
			usleep( 1000*tries );
			ptyxm_fd = open( fixed_ttyname, O_RDWR | O_NONBLOCK);
		}
		if ( ptyxm_fd >= 0 )
		{
			succeeded=1;
			len = 1;
			ret = ioctl( ptyxm_fd, TIOCPKT, &len );
			if ( ret )
			{
				trace( tl_error, "Error in ioctl-1: %d %d\n", ret, errno );
				exit( -1 );
			}
			ptyx_mode = 1;

			trace( tl_debug, "Opened port %s \n", fixed_ttyname );
			return ptyxm_fd;
		}
		else
		{
			trace( tl_error, "Error in opening %s errno=%d\n", fixed_ttyname, errno  );
			trace( tl_error, "   - %s\n", strerror( errno ) );
			if (succeeded)
			{
				return -1;
			}
			else
			{
				exit(-1);
			}
		}
	}
	else
		ptyx_mode = 0;
	for( i = 0; i < 256; i++ )
	{
		sprintf( f_name, "/dev/tx%d", i );
		ptyxm_fd = open( f_name, O_RDWR | O_NONBLOCK );
		if ( ptyxm_fd >= 0 )
		{
			ret = ioctl( ptyxm_fd, TIOCPKT, &len );
			if ( ret )
			{
				trace( tl_error, "Error in ioctl-2: %d %d\n", ret, errno );
				exit( -1 );
			}

			return ptyxm_fd;
		}

	}

	if ( ptyxm_fd < 0 )
	{
		ptyxm_fd = open( "/dev/ptmx", O_RDWR | O_NONBLOCK );
		if ( ptyxm_fd < 0 )
		{
			trace( tl_error, "Error opening /dev/ptmx\n" );
			return -1;
		}

		grantpt( ptyxm_fd );  /* change slave permissions */
		unlockpt( ptyxm_fd );  /* unlock slave */
		ptydev = ptsname( ptyxm_fd ); /* get slave's name */

		/* The linkname ends up with a double "/". This is not a
		   bug. This allows us to recognize our own links with higher
		   confidence than we would otherwise have. -- REW */

		sprintf( linkname, "/dev/%s", ptydev+4 );
		ptydev = linkname;
	}

	if ( pty_mode != -1 )
		chmod( ptydev, pty_mode );

	/* chown will not modify anything if either uid or gid is -1... */
	chown( ptydev, pty_uid, pty_gid );


	fcntl( ptyxm_fd, F_SETFL, fcntl(ptyxm_fd,F_GETFL) | O_NONBLOCK );

	tcgetattr( ptyxm_fd, &modes );
	modes.c_lflag &= ~ECHO;
	tcsetattr( ptyxm_fd, TCSANOW, &modes );

	if ( fixed_ttyname )
	{
		unlink( fixed_ttyname );
		symlink( ptydev, fixed_ttyname );

		fixed_dir = strdup( fixed_ttyname );
	}
	else
		fixed_dir = strdup( ptydev );

	p = strrchr( fixed_dir, '/' );
	if ( p )
		*p = 0;
	else
	{
		free( fixed_dir );
		fixed_dir = strdup( "/dev/" );
	}

	chdir( fixed_dir );
	dirp = opendir( fixed_dir );
	free( fixed_dir );

	if ( dirp )
	{
		while( (dp = readdir( dirp )) )
		{
			lstat( dp->d_name, &stat_buf );
			if ( S_ISLNK( stat_buf.st_mode ) )
			{
				len = readlink( dp->d_name, buf, 1024 );
				buf[len] = '\0';
				if ( !strcmp( ptydev, buf ) )
				{
					unlink( dp->d_name );
					symlink( ptydev, fixed_ttyname );
				}
			}
		}
		closedir( dirp );
	}
	else
	{
		/* Couldn't open /dev . Well then we can't verify if there
		   are other links to our device. Sorry about that....
		   But this shouldn't be fatal, and we continue anyway. -- REW */
		trace( tl_error, "Warning: Can't opendir /dev/\n" );
	}
	len = 1;

	/* This will most likely fail. Ignore the error. -- REW */
	symlink( ptydev, fixed_ttyname );
	ret = ioctl( ptyxm_fd, TIOCPKT, &len );
	if ( ret )
	{
		trace( tl_error, "Error in ioctl-3: %d %d\n", ret, errno );
		exit( -1 );
	}
	return ptyxm_fd;
}


void close_tty(void)
{
	trace( tl_debug, "close_tty(): Closeing ptyxm_fd=%d\n", ptyxm_fd );

	close_tty_common(1);
}


void close_tty_common(int rmdev)
{
	if (ptyxm_fd != -1)
	{
		close(ptyxm_fd);
		if (rmdev && !ptyx_mode)
		{
			unlink(fixed_ttyname);
		}
		tty_connected = 0;
		ptyxm_fd = -1;
		slave_is_open = FALSE;
		tty_write_blocked=0;
		pending_drain=0;
	}
}

void close_tty_no_rm(void)
{
	close_tty_common(0);
}


/****************************************************************************
 *	Get a message name from a message list		    		    *
 ****************************************************************************/
/*
	Arguments	: msg_val  : value of message to look up
			  msg_list : list of value/name
	Return value	: pointer to name
*/
char *get_msg(int msg_val, msg_name_t *msg_list)
{
	int j;

	for (j=0; msg_list [j].msg != msg_val && msg_list [j].msg != -1; j++)
		;
	return msg_list [j].name;
}


/****************************************************************************
 *	Close connections						 				    		    *
 ****************************************************************************/
/*
	Arguments	: - None
	Return value	: - None
*/
void close_connection()
{
	/* close */
	trace(tl_debug, "closing device server connection");

	/* tidy up */
	if (dserver_fd != -1)
	{
		// try to forward any held frames
		forward_frame(&info );
		if ( close(dserver_fd) < 0 )
		{
			 trace(tl_debug, "close_connection(): Error closing dserver_fd=%d, errno: %d",dserver_fd, errno);
		}
		if (ptyxm_fd != -1)
		{
			if (open_wait_time == OPENWAIT_ALWAYS_SUCCESSFUL)
			{
				drop_modem_signals();
			}
			tp_connected = FALSE;
			write_tty_ctrl(ptyxm_fd, TP_PROTO_TP_CONNECTION, &tp_connected, 1);
		}
		keepalive_sent = 0;
	}
	dserver_fd = -1;
	if (udp_sock != -1)
	{
		close(udp_sock);
	}
	udp_sock = -1;
	dserver_connected = FALSE;
	tp_connected = FALSE;
	signal_state_sent = 0;
	flushing_net_read = 0;	// clean up in case we didn't get the DS_TCP_SYNC_FLUSH_END from the server yet
	DServerFeatures = 0;
	// as long as we are not blocked on writting to the tty then clear out read data
	// from network
	if (!tty_write_blocked)
	{
		trace(tl_debug, "close_connection(): tty write not blocked so throwing away %d bytes rx", dserver_in_buf.len );
		buf_init(&dserver_in_buf);
	}
	if ( close_tty_flag && (ptyxm_fd != -1) )
	{
		trace(tl_info, "Closing tty connections\n");
		close_tty();
	}


#ifdef	USE_SSL
	if ( ssl )
	{
		SSL_free( ssl );
		ssl = NULL;
	}
#endif	//USE_SSL
}


//******************************************************************************
static	void	close_connection_2( int fd )
{
	dserver_fd = fd;
	close_connection();
}


static void my_hd (void *ad, int len)
{
	unsigned char *addr = ad;
	int i, j, ch;
	char			buf[256];

	buf[0] = 0;
	for (i=0;i<len;i+=16)
	{
		sprintf ( &buf[strlen(buf)], "%p ", (void *)addr+i);
		for (j=0;j<16;j++)
		{
			sprintf (&buf[strlen(buf)], "%02x %s", addr[j+i], (j==7)?" ":"");
		}
		for (j=0;j<16;j++)
		{
			ch = addr[j+i];
			sprintf (&buf[strlen(buf)], "%c", (ch < 0x20)?'.':((ch > 0x7f)?'.':ch));
		}
		sprintf (&buf[strlen(buf)], "\n");
	}
}


//******************************************************************************

void    trace(trace_level_t level, char *msg, ...)
{
	va_list ap;
	char    *arg [7];
	char    buf [2049];
	time_t  now;
	struct tm *tim;

	/* provide fast exit if we're not going to do anything */

	if (level > trace_level)
		return;

	va_start(ap, msg);

	arg [0] = va_arg(ap, char *);
	arg [1] = va_arg(ap, char *);
	arg [2] = va_arg(ap, char *);
	arg [3] = va_arg(ap, char *);
	arg [4] = va_arg(ap, char *);
	arg [5] = va_arg(ap, char *);
	arg [6] = va_arg(ap, char *);

	va_end(ap);

	sprintf(buf, msg, arg [0], arg [1], arg [2], arg [3], arg [4], arg [5], arg [6]);

	if (buf[strlen(buf)-1] != '\n')
		strcat(buf, "\n");

	time(&now);
	tim = localtime(&now);

	if (trace_file)
	{
		fprintf(trace_file, "%02d:%02d:%02d %s: %s",
		         tim->tm_hour, tim->tm_min, tim->tm_sec,
		         trace_level_str [level], buf);
		fflush(trace_file);
		fprintf( trace_file, " \b" );
	}
	else
	{
		/* Useful when debuging only !*/
		fprintf(stderr, "COMredirect: %s", buf);
	}
}



//******************************************************************************
// returns:	network socket fd for connection
//				< 0 on error
//
int wait_for_dserver_connection(void)
{
	int 	nsfd = -1;
	int 	on = 1;
	int     tries;
	int	ret;
	int	force_try=0;
    int doretry=0;
	
	// Configured to initiate TCP connection
	if (client_tcp_host)
	{
		 // if necessary wait until the appliction has opened or reopened the slave tty
		 // if connect_on_init is enabled, don't bother waiting for the app to open
		 // the serial port before attempting to make the TCP connection for 1 time
		 // only
		if (connect_on_init)
		{
			connect_on_init=0;
			force_try=1;
			doretry=1;
		}
		else if (netfailed)
		{
			netfailed = 0;
			if (restorenet)
			{
				force_try=1;
				doretry=1;
			}
		}
		else if (tty_connected && !slave_is_open)
		{
			trace(tl_debug, "Client Initiated needs slave tty open to start connection");
			ret = wait_for_tty_open_close(NULL);
			if( (ret == -1) || !tty_connected)
				return(FAILURE);

			force_try=1;
			doretry=1;
		}
		trace(tl_debug, "%s: About try to connect, retry_num=%d",__FUNCTION__,retry_num);
		// try to connect for configured retry number and time
		for (tries=0; force_try || (doretry && (retry_num == -1 || tries < retry_num)) ;)
		{
			// Try to connect to device server
			if ( (nsfd = wait_for_tcp_connect()) >= 0 )
			{
#ifdef USE_SSL
				if ( sslcfg.enabled )
				{
					if ( wait_for_ssl_connect(nsfd) >= 0)
					{
						break;		// connected, break out of retries
					}
				}
				else
#endif //USE_SSL
				{
					break; // connected, break out of retries
				}
			}
			nsfd = -1;			// nsfd is invalid now
			trace(tl_status, "wait_for_tcp_connect() failed tries=%d", tries);

			if (!force_try)
			{
				 tries++;
			}
			force_try=0;
			if (retry_num==-1 || tries < retry_num)
			{
				 // wait for retry time before trying again 
				 sleep(retry_time);
			}
		}  // end for ()

		// if restore network connection is disabled or retry has expired, return connection failed
		if (nsfd < 0 && (!doretry || retry_num >= 0))
		{
			// main line will reopen the tty
			trace(tl_status, "wait_for_dserver_connection(): Connect retry expired or no restore connection , closing tty");
			drop_modem_signals();
			close_tty();
			return(-1);
		}
		netfailed=1;	// default to network failed.  this is set to 0 at specific code locations, when we close the network connection
	}
	// Configured to listen for TCP connection
	else
	{
		nsfd = wait_for_tcp_listen();
		if ( nsfd < 0 )
		{
			return(-1);
		}
#ifdef USE_SSL
		if ( sslcfg.enabled )
		{
			if ( wait_for_ssl_connect(nsfd) < 0)
			{
				 return(-1);
			}
		}
#endif //USE_SSL
	}

	// ********** Common already TCP Connected logic *****************
	trace(tl_status, "Connected to Device Server(fd=%d)", nsfd);
	dserver_connected = TRUE;


	setsockopt(nsfd, SOL_SOCKET, SO_KEEPALIVE, (char *)&on, sizeof(on));

	if (mode == pm_lite)
	{
		setsockopt(nsfd, SOL_SOCKET, SO_SNDBUF,
					 (const char *) &send_bufsize, sizeof(send_bufsize));
	}
	if (no_tcp_nagle)
	{
		 setsockopt(nsfd, IPPROTO_TCP, TCP_NODELAY, (char *)&on, sizeof(on));
	}

	fcntl(nsfd, F_SETFL, fcntl( nsfd, F_GETFL ) | O_NONBLOCK );

	memset( &info, 0, sizeof( forwarding_info_t )  );
	if (keepalive_time)
	{
		gettimeofday( &info.keepalive_start, NULL );
	}
	if ( pkt_fwd_enabled || opmode_type == OPMODE_TYPE_CUSTOM )
	{
		if ( get_pkt_fwd_config( &info, tcp_host, tcp_port ) < 0  )
		{
			trace( tl_error, "Packet forwarding setup failed.\n" );
			close_connection_2(nsfd);
			return -1;
		}
	}
	else
	{
/* now let's try to fix up the config taking into account the operation mode
*/
		if (opmode_type == OPMODE_TYPE_PACKETIDLETIMEOUT)
		{
			info.idle_time = idle_time_packet_timeout;
			trace( tl_debug, "Operation mode: %s, idle timeout=%d.\n",opmode_name, idle_time_packet_timeout);
			if (info.idle_time)
			{
				info.forwarding |= PKT_FORWARD_ON_IDLE;
                info.char_buf_ptr = &info.char_buf[0];
			}
		}
	}
	if (is_tp_conn_avail()) 
	{
		// update driver that connection is up
		write_tty_ctrl(ptyxm_fd, TP_PROTO_TP_CONNECTION, &tp_connected, 1);

		// if in Lite mode then fake out modem signals and send TP_PROTO_DCD_UP message
		// to unblock any waiting opens since TCP connection is now up
		if (mode == pm_lite)
		{
			ret = ioctl(ptyxm_fd, TIOCMGET, &current_state);	// get modem status
			if (ret)
			{
				trace(tl_error, "%s(): TIOCMGET error %d %d\n", __FUNCTION__, ret, errno);
				close_connection_2(nsfd);
				return(-1);
			}
			current_state|=(TIOCM_DSR|TIOCM_CD|TIOCM_CTS);
			ret = ioctl(ptyxm_fd, PTX_IOCMSET, &current_state);	// set modem status
			if (ret)
			{
				trace(tl_error, "%s(): PTX_IOCMSET error %d %d\n", __FUNCTION__, ret, errno);
				close_connection_2(nsfd);
				return(-1);
			}
		}
	}
	return nsfd;                /* Return new socket */
}


//******************************************************************************
// wait_for_tcp_connect
// returns:	network socket fd for connection
//				< 0 on error
//
int wait_for_tcp_connect(void)
{
	struct addrinfo hints; 
	struct addrinfo *res = NULL;
	int nsfd = -1;
	int rc;
	char portstr[7];
	struct	sockaddr_in6 saddr6;
	struct sockaddr_in saddr;
	struct linger linger;
	int syn_count = 3;

	ipv6_not_running = 0;  // clear ipv6 flag
   memset((char *) &hints, 0, sizeof(struct addrinfo));    // Clear out hints

	trace(tl_debug, "wait_for_tcp_connect() Enter ");
	
	do
	{
		hints.ai_flags    = 0;
		hints.ai_family   = AF_UNSPEC;
		hints.ai_socktype = SOCK_STREAM;
		// Convert the text form of the address to binary
		// form. If it is numeric then we want to prevent getaddrinfo()
		// from doing any name resolution.
		memset((char *) &saddr, 0, sizeof(saddr));    // Clear address structure
		rc = inet_pton(AF_INET, client_tcp_host, &saddr);
		if (rc == 1)				// valid IPv4 text address ?
		{
			hints.ai_family = AF_INET;
			hints.ai_flags |= AI_NUMERICHOST;
			ipv6_not_running = 1;
		}
		else
		{
   		memset((char *) &saddr6, 0, sizeof(saddr6));    // Clear address structure
			rc = inet_pton(AF_INET6, client_tcp_host, &saddr6);
			if (errno == EAFNOSUPPORT)
			{
				ipv6_not_running = 1;
			}
			else if (rc == 1) /* valid IPv6 text address? */
			{
				hints.ai_family = AF_INET6;
				hints.ai_flags |= AI_NUMERICHOST;
			}
		}
		// Get the address information for the device server
		snprintf(portstr, 6, "%d", tcp_port);			// convert tcp_port to string
		rc = getaddrinfo(client_tcp_host, portstr, &hints, &res);
		if (rc != 0)
		{
			trace(tl_error, "Host name %s could not be found:  %s ", 
							client_tcp_host, gai_strerror(errno));
			if (rc == EAI_SYSTEM)
			{
				trace(tl_error, "getaddrinfo() failed"); 
			}
			break;
		}
		
		nsfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if (nsfd < 0) 
		{
			trace(tl_error, "can't open socket: %s", strerror(errno));
			break;
		}
		setsockopt(nsfd, SOL_TCP, TCP_SYNCNT, &syn_count, sizeof(syn_count));
		// use host address information to try to connect

		trace (tl_debug, "calling connect() - fd=%d\n", nsfd);

		rc = connect(nsfd, res->ai_addr, res->ai_addrlen);

		if (rc < 0)
		{
		  	close(nsfd);
			nsfd = -1;
			trace(tl_error, "connect() failed: %s\n", strerror(errno));
			break;
		}
		// make sure all data is sent out on TCP close
		linger.l_onoff = 1; 
		linger.l_linger = 5;
		setsockopt(nsfd, SOL_SOCKET, SO_LINGER, &linger, sizeof(linger));
		
		// now we are connected fill in dserver address info
		if (res->ai_family == AF_INET6)
		{
	 		memset(&dserver_addr6.sin6_addr, 0, sizeof (struct in6_addr));
			dserver_addr6.sin6_family = AF_INET6;
			memcpy(&dserver_addr6, (struct sockaddr_in *)res->ai_addr, res->ai_addrlen);
			
		}
		else if (res->ai_family == AF_INET)
		{
			ipv6_not_running = 1;
 	 		memset(&dserver_addr.sin_addr, 0, sizeof (struct in_addr));
			dserver_addr.sin_family = AF_INET;
			memcpy(&dserver_addr, (struct sockaddr_in *)res->ai_addr, res->ai_addrlen);

		}
		else
		{
			// should never get here
			trace(tl_error, "wait_for_tcp_connect() failed: Invalid Address Family \n", res->ai_family);
			break;
		}
	} while(FALSE);

	// common clean-up
	if (res != NULL)
		freeaddrinfo(res);
	return(nsfd);
}


//******************************************************************************
// returns:	network socket fd for connection
//				< 0 on error
//
int wait_for_tcp_listen(void)
{
	int 		on = 1;
	struct 	sockaddr_in 	saddr;
	socklen_t				saddrlen = sizeof(struct sockaddr_in);		// Temp for addr length
	int 					listenfd = -1;
	int i, rc, rc2;
	struct addrinfo hints; 
	struct addrinfo *res = NULL;
	int nsfd = -1;
	int closensfd = 0;
	struct sockaddr *sa;
	int	salen;
	char ipstr[INET6_ADDRSTRLEN];
	struct	sockaddr_in6	saddr6;
	socklen_t saddr6len = sizeof(struct sockaddr_in6);	// Temp for addr length

	ipv6_not_running = 0;  // clear ipv6 flag
   memset((char *) &hints, 0, sizeof(struct addrinfo));    // Clear out hints
	
	do
	{
		// Resolve server host name/IP address if specified 
		if (server_tcp_host)
		{
	
			hints.ai_flags    = 0;
			hints.ai_family   = AF_UNSPEC;
			hints.ai_socktype = SOCK_STREAM;
			// Convert the text form of the address to binary
			// form. If it is numeric then we want to prevent getaddrinfo()
			// from doing any name resolution.  
			memset((char *) &saddr, 0, sizeof(saddr));    // Clear address structure
			rc = inet_pton(AF_INET, server_tcp_host, &saddr);
			if (rc == 1)				// valid IPv4 text address ?
			{
				hints.ai_family = AF_INET;
				hints.ai_flags |= AI_NUMERICHOST;
				ipv6_not_running = 1;
			}
			else
			{
	 			memset((char *) &saddr6, 0, sizeof(saddr6));     /* Clear address structure */
				rc = inet_pton(AF_INET6, server_tcp_host, &saddr6);
				if (errno == EAFNOSUPPORT)
				{
					 ipv6_not_running = 1;
				}
				else if (rc == 1) /* valid IPv6 text address? */
				{
					 hints.ai_family = AF_INET6;
					 hints.ai_flags |= AI_NUMERICHOST;
				}
			}
			// Get the address information for the configured host
			rc = getaddrinfo(server_tcp_host, NULL, &hints, &res);
			if (rc != 0)
			{
				trace(tl_error, "Host name %s could not be found:  %s ", 
								server_tcp_host, gai_strerror(errno));
				if (rc == EAI_SYSTEM)
				{
					trace(tl_error, "getaddrinfo() failed"); 
				}
				break;
			}
			sa = res->ai_addr;
			salen = res->ai_addrlen;
			if (res->ai_family == AF_INET6)
			{
				if ( inet_ntop(AF_INET6, &((struct sockaddr_in6 *)sa)->sin6_addr, ipstr, sizeof(ipstr)) )
				{
					trace(tl_info, "Configured host name %s resolved to IP address %s ", server_tcp_host, ipstr);
				}
			}
			else if (res->ai_family == AF_INET)
			{
				ipv6_not_running = 1;
				if ( inet_ntop(AF_INET, &((struct sockaddr_in *)sa)->sin_addr, ipstr, sizeof(ipstr)) )
				{
				   trace(tl_info, "Configured host name %s resolved to IP address %s ", server_tcp_host, ipstr);
				}
			}
			else
			{
				trace(tl_info, "getaddrinfo() failed: Unsupported INET family ");
				break;
			}
		}
		
		listenfd = socket(PF_INET6, SOCK_STREAM, 0);  /* Go open socket */
		
		// if IPv6 protocol not running/available then use IPv4
		if (ipv6_not_running || (listenfd == -1 && errno == EAFNOSUPPORT) )
		{
			if ((listenfd = socket(PF_INET, SOCK_STREAM, 0)) < 0) /* Go open socket */
			{
				trace(tl_error, "can't open socket: %s", strerror(errno));
				break;
			}
			ipv6_not_running = 1;
			memset((char *) &saddr, 0, sizeof(saddr));    // Clear address structure
			saddr.sin_family = AF_INET;         			// Set family
			saddr.sin_addr.s_addr = INADDR_ANY;
			saddr.sin_port = htons((u_short) tcp_port);  	// .. and listener port
			setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
			if (bind(listenfd,(struct sockaddr *)&saddr, sizeof(saddr)) < 0)  // Go bind address to socket
			{
				trace(tl_error, "can't bind on listen address: %s", strerror(errno));
				break;
			}
	
			trace (tl_debug, "calling listen() - fd=%d\n", listenfd);
			while ((i = listen(listenfd, 1)) < 0)       /* Wait for a connection */
			{
				trace(tl_error, "listen failed: %s\n", strerror(errno));
				sleep(5);
			}
	
			trace (tl_debug, "calling accept\n");
			if ((nsfd = accept(listenfd,(struct sockaddr *) &dserver_addr, &saddrlen)) < 0) /* Accept caller */
			{
				trace(tl_error, "can't accept connection: %s", strerror(errno));  // Failed, clean-up below
				break;
			}
		}
		else if ( listenfd == -1 )
		{
			trace(tl_error, "can't open socket(6): %s (%d)", strerror(errno), errno );
			break;
		}
		else
		{
			memset((char *) &saddr6, 0, sizeof(saddr6));	/* Clear address structure */
			saddr6.sin6_family = AF_INET6;        			 	/* Set family */
			saddr6.sin6_port = htons((u_short) tcp_port);  /* .. and listener port */
			setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
			if (bind(listenfd,(struct sockaddr *)&saddr6, sizeof(saddr6)) < 0) /* Go bind address to socket */
			{
				trace(tl_error, "can't bind on listen address(6): %s", strerror(errno));
				break;
			}
	
			trace (tl_debug, "calling listen() - fd=%d\n", listenfd);
			while ((i = listen(listenfd, 1)) < 0)       /* Wait for a connection */
			{
				trace(tl_error, "listen failed(6): %s\n", strerror(errno));
				sleep(5);
			}
	
			trace (tl_debug, "calling accept\n");
			if ((nsfd = accept(listenfd,(struct sockaddr *) &dserver_addr6, &saddr6len)) < 0) /* Accept caller */
			{
				trace(tl_error, "can't accept connection(6): %s", strerror(errno)); /* Failed */
				break;
			}
		}
		// Validate remote host if required
		if (server_tcp_host)
		{
			// get address of peer and compare to configured host
			if (ipv6_not_running)
			{
				rc = getpeername(nsfd, (struct sockaddr *)&dserver_addr, &saddrlen);
				if (rc == 0)
				{
					rc2 = memcmp( &((struct sockaddr_in *)sa)->sin_addr, &dserver_addr.sin_addr, 4);
					if ( inet_ntop(AF_INET, (struct sockaddr_in *)&dserver_addr.sin_addr, ipstr, sizeof(ipstr)) )
					{
						trace(tl_info, "Peer host IP4 address is %s ", ipstr);
					}
				}
			}
			else
			{
				rc = getpeername(nsfd, (struct sockaddr *)&dserver_addr6, &saddr6len);
				if (rc == 0)
				{
					 rc2 = memcmp(&((struct sockaddr_in6 *)sa)->sin6_addr, &dserver_addr6.sin6_addr, 16);
					if ( inet_ntop(AF_INET6, (struct sockaddr_in6 *)&dserver_addr6.sin6_addr, ipstr, sizeof(ipstr)) ) 
					{
						trace(tl_info, "Peer host IP address is %s ", ipstr);
					}
				}
			}
			if (rc < 0)
			{
				trace(tl_error, "getpeername() failed: %s (%d)", strerror(errno), errno);
				closensfd = 1;
				break;
			}
			if (rc2 != 0)
			{
				trace(tl_error, "Peer host does not match configured server host %s ", server_tcp_host);
				closensfd = 1;
				break;
			}
		} // end validate server_tcp_host
		
	} while(FALSE);

	// common clean-up
	if (listenfd != -1)
		close(listenfd);
	if (closensfd)
	{
		 close(nsfd);
		 nsfd = -1;
	}
	if (res != NULL)
		freeaddrinfo(res);
	return(nsfd);
}


//******************************************************************************
// returns:	< 0 on error, including timeout
//
#ifdef USE_SSL
int wait_for_ssl_connect(int nsfd)
{
	int rc, ret;
	fd_set read_set;
	fd_set write_set;
	struct timeval	tim;
	int ssl_error;
	const char *err_string;

	if ( get_ssl_config(tcp_host, tcp_port) != 0 )
	{
		trace( tl_error, "get_ssl_config failed\n" );
		close_connection_2(nsfd);
		return FAILURE;          /* Return failure */
	}

	trace( tl_info, "initializing SSL...\n" );
	if ( ssl_init( &ssl_ctx, tp_pem_passwd_cb, tp_tmp_rsa_cb ) < 0 )
	{
		close_connection_2(nsfd);
		return FAILURE;          /* Return failure */
	}
	ssl = SSL_new( ssl_ctx );
	SSL_set_fd( ssl, nsfd );

#if	0	//debug only
	{
		int 	i = 0;
		char	*cipher;
		char	cipherList[2048];
		memset( cipherList, 0, 2048 );
		while ( (cipher = SSL_get_cipher_list( ssl, i )) != NULL )
		{
			trace( tl_info, "%d: %s\n", i, cipher );
			i++;
		}
	}
#endif
	// If device server is configured as an SSL server then accept incoming
	//  SSL connections
	if ( sslcfg.ssl_type == SSL_SERVER )
	{

		trace( tl_info, "waiting for SSL accept\n" );
		while(1)
		{
			if ( (rc = SSL_accept( ssl )) > 0 )
				break;		//successful

			switch( ssl_error = SSL_get_error( ssl, rc ) )
			{

				case SSL_ERROR_WANT_WRITE:
				case SSL_ERROR_WANT_READ:
				FD_ZERO( &read_set );
				FD_ZERO( &write_set );
				if ( ssl_error == SSL_ERROR_WANT_READ )
					FD_SET( nsfd, &read_set );
				else
					FD_SET( nsfd, &write_set );

				tim.tv_sec  = SSL_ACCEPT_OR_CONNECT_TIMEOUT_SECS;
				tim.tv_usec = 0;
				ret = select(FD_SETSIZE, &read_set, &write_set, NULL, &tim );

				if ( ret == 0)
				{
					tim.tv_sec  = 0;
					tim.tv_usec = 0;
					
					ret = select(FD_SETSIZE, &read_set, &write_set, NULL, &tim);

					if ( ret <= 0 )
					{
						trace( tl_error, "SSL_accept failed - timeout\n" );
						close_connection_2(nsfd);
						return FAILURE;          /* Return failure */
					}
					break;		//
				}
				else if ( ret < 0 )
				{
					if((err_string = strerror(errno)) == NULL )
						err_string = "...";
					trace( tl_error, "SSL_accept failed - %s", err_string );
					close_connection_2(nsfd);
					return FAILURE;          /* Return failure */
				}
				else
					trace( tl_debug, "%s event occured.", "SSL_accept" );
				// keep trying while getting events from select
				break;

				default:
				if((err_string = ERR_reason_error_string( ERR_get_error() )) == NULL )
				{
					err_string = "...";
				}
				trace( tl_status, "%s", err_string );
				trace( tl_error, "SSL_accept failed - error = %x", ssl_error );
				close_connection_2( nsfd );
				return FAILURE;          /* Return failure */
				break;
			}
		}
	}

	// device server is configured as an SSL client. Make an SSL connection.
	else
	{
		trace( tl_info, "waiting for SSL connect\n" );
		while(1)
		{
			if ( (rc = SSL_connect( ssl )) > 0 )
				break;		//successful

			switch( ssl_error = SSL_get_error( ssl, rc ) )
			{

				case SSL_ERROR_WANT_WRITE:
				case SSL_ERROR_WANT_READ:
				FD_ZERO( &read_set );
				FD_ZERO( &write_set );
				if ( ssl_error == SSL_ERROR_WANT_READ )
					FD_SET( nsfd, &read_set );
				else
					FD_SET( nsfd, &write_set );

				tim.tv_sec  = SSL_ACCEPT_OR_CONNECT_TIMEOUT_SECS;
				tim.tv_usec = 0;

				ret = select(FD_SETSIZE, &read_set, &write_set, NULL, &tim );
				
				if ( ret == 0)
				{
					tim.tv_sec  = 0;
					tim.tv_usec = 0;
					ret = select(FD_SETSIZE, &read_set, &write_set, NULL, &tim);
					
					if ( ret <= 0 )
					{
						trace( tl_error, "SSL_connect failed - timeout\n" );
						close_connection_2(nsfd);
						return FAILURE;          /* Return failure */
					}
					trace( tl_debug, "%s event occured after timeout.",  "SSL_connect" );
					break;		//
				}
				else if ( ret < 0 )
				{
					if((err_string = strerror(errno)) == NULL )
						err_string = "..";
					trace( tl_error, "SSL_connect failed - %s\n", err_string );
					close_connection_2(nsfd);
					return FAILURE;          /* Return failure */
				}
				else
					trace( tl_debug, "%s event occured.",  "SSL_connect" );
				// keep trying while getting events from select
				break;

				default:
				//					ERR_print_errors( bio_err );		//Debug
				if((err_string = ERR_reason_error_string( ERR_get_error() )) == NULL )
					err_string = "...";
				trace( tl_status, "%s\n", err_string );
				trace( tl_error, "SSL_connect failed - error = %x\n", ssl_error );
				close_connection_2(nsfd);
				return FAILURE;          /* Return failure */
				break;
			}
		}
	}

	// authenticate the peer.
	if ( sslcfg.do_authentication )
	{
		trace( tl_info, "Authenticating peer(%s)\n",
			   (sslcfg.ssl_type == SSL_SERVER ? "client" : "server") );
		if ( !check_cert( ssl ) )
		{
			trace( tl_error, "Peer authentication failed\n" );
			close_connection_2(nsfd);
			return FAILURE;          /* Return failure */
		}
	}
	trace( tl_info, "SSL connection complete!\n" );
	trace( tl_info, "Protocol = %s, Cipher = %s\n",
		   SSL_get_cipher_version( ssl ),
		   SSL_get_cipher_name( ssl ) );
	return(1);
}
#endif // USE_SSL




void dump_buf(FILE *fd, unsigned char * buf, int len)
{
	int j, k;
	if(len > MAX_TRACE_LEN)
		len = MAX_TRACE_LEN;
	for(j=0; j<len; j+=16)
	{
		fprintf(fd, "%04x ", j);
		for(k=0; k<16; k++)
			if(j+k < len)
				fprintf(fd, "%02x ",(unsigned char)buf [j+k]);
			else
				fprintf(fd, "   ");
		fprintf(fd, "  ");
		for(k=0; k<16 && j+k<len; k++)
			if(isprint(buf [j+k]))
				fprintf(fd, "%c", buf[j+k]);
			else
				fprintf(fd, ".");
		fprintf(fd, "\n");
	}
}



//******************************************************************************
const char	*my_basename(const char *str)
{
	const char     *base = strrchr(str, '/');

	return(base ? base + 1 : str);
}




/****************************************************************************
 *	Write data to a file descriptor
 ****************************************************************************/
/*
	Arguments	: fd   : descriptor of file
			  buf  : pointer to data buffer
			  size : number of bytes
	Return values	: -1 if error(errno is set), 
					or 0/+Ve number of bytes(errno EAGAIN),
					otherwise no of bytes written
*/
int write_buf_raw( int fd, unsigned char *buf, int size )
{
	int  len;
	unsigned char *cp;
	int  ret;
	
#ifdef USE_SSL
	int	 err;
#endif

	cp = buf;
	len = size;

	// if dserver not connected, and this includes SSL if configured
	// then return error
	if( (dserver_connected == FALSE) || (fd == -1) )
		return FAILURE;
  
#ifdef USE_SSL
	if( sslcfg.enabled )
	{
		do
		{
			ret = SSL_write( ssl, cp, len );

			switch((err = SSL_get_error( ssl, ret )) )
			{

				case SSL_ERROR_NONE:
				cp += ret;
				len -= ret;
				if( len <= 0 )
				{
					write_ssl = 0;
				}
				else /* if (len > 0) */
					write_ssl=1;

				break;

				case SSL_ERROR_WANT_WRITE:
				trace (tl_data, "SSL_write returns %d: %s\n", ret, strerror (errno) );
				write_ssl=1;
//				return( size - len);
				ret = size - len;
				goto write_buf_raw_done;

			default:
				trace(tl_error, "SSL_write error %d: %s\n", ret, strerror(errno) );
				if( err == SSL_ERROR_ZERO_RETURN && ret == 0 )
				{
//					return ret;
					goto write_buf_raw_done;
				}
				else
				{
//					return -1;
					ret = -1;
					goto write_buf_raw_done;
				}
			}
			if( len > 0 )
				trace(tl_debug, "partial SSL_writer - looping" );
		}
		while ( len > 0 );
	}
	else
#endif  // USE_SSL
	{
		do
		{
			ret = write(fd, cp, len );
			if ( ret < 0 )
			{
				if (errno == EAGAIN)
				{
					trace (tl_data, "write_buf_raw(): write(fd=%d) returned %d: %s, bytes written=%d\n", 
									fd, ret, strerror (errno), size-len );
					/* Return count of bytes written */
//					return (size-len);
					ret = size - len;
					goto write_buf_raw_done;
				}
				else
				{
					trace(tl_error, "write_buf_raw(): write_dserver error %d: %s\n", ret, strerror(errno) );
					/* Error condition so return the error */
//					return ret;
					goto write_buf_raw_done;
				}
			}
			else if(ret < len)
			{
				trace(tl_debug, "write_buf_raw(): partial write_dserver - looping" );
			}
			
			if(trace_level >= tl_data)
			{
				trace(tl_data, "write_buf_raw(): write_device_server %d bytes",ret);
				dump_buf(trace_file, cp, ret);
			}
			
			len -= ret;
			cp += ret;
		}
		while ( len > 0 );
	}
	/* Wrote everything ! */
	ret = size;
	
write_buf_raw_done:
	 
	// restart keep-alive timer if in Lite mode
	// resolves issue with tcp zero-window probe response being lost when
	// urgent data message is sent
	if ( (keepalive_time) && (ret > 0) && (mode == pm_lite) )
	{
		gettimeofday( &info.keepalive_start, NULL );
	}
	return( ret );
}

/****************************************************************************
 *	Read from an DEVICE_SERVER connection
 ****************************************************************************/
/*
	Arguments	: fd   : descriptor of socket
			  type : type of packet read
			  buf  : pointer to data buffer
	Return values	: >0   : ok, number of bytes read
			   0   : connection closed
			  -1   : if error(errno may be set)
*/
int read_dserver(int fd, dserver_packet_t *type, buf_t * buf)
{
	int  ret;
	unsigned short temp;
	static  short header, count = 0;
	unsigned char new_state;
	unsigned char temp_state;
	
	//****************************
	//   Lite Mode
	if (mode == pm_lite)
	{
		{
			*type = DEVICE_SERVER_DATA;
			ret = tp_read(fd, (unsigned char *)buf->data, sizeof(buf->data));
			if (ret > 0)
				buf->len = ret;
			if(trace_level >= tl_data)
			{
				trace(tl_data, "read_dserver(): Lite mode tp_read returned %d", ret);
				dump_buf(trace_file, buf->data, buf->len);
			}
			return ret;
		}
	}

	//***************************
	//   Packet Mode(Full)
	else
	{
		if(state == s_sent_packet_req)
		{
			ret = tp_read(fd, (unsigned char *)buf->data, 1);
			if (ret==1 && (buf->data[0] == DS_URG_TCP_PACKETMODE_ACK))
			{
				trace(tl_info, "received packet mode ack\n");
				*type = DS_URG_TCP_PACKETMODE_ACK;
				state = s_wait_port_number;
				return 1;
			}
			else
			{
				// Do not cause error if data not correct for packet mode ack.
				// There is now a window where data can be received from the device
				//	server when SSL has been configured. Throw out data until packet
				//	mode ack is received.
				trace(tl_data, "packet mode ack not received yet - ret = %d, data = %x\n", ret, buf->data[0] );
				*type = DEVICE_SERVER_NONE;  // This will cause the data to be ignored at the upper level.(i.e. throw out the data)
				return ret;
			}
		}
		else
		{
			if(buf->len == 0)
			{
				/* if in buffer is empty, assume we read new packet */
				ret = tp_read( fd,(unsigned char *)&temp, 2 );
				if(ret <= 0 )
				{
					return ret;
				}
				
				// If we can't read two bytes here we are obviously messed up
				// 
//				assert(ret==2);
				if (ret !=2)
				{
					trace(tl_data, "read_dserver(): Expecting tp_read to return 2 bytes but read %d, fake out and return 0 bytes read", 	ret);
					return 0;
				}

				header = ntohs(temp);

				count = header & DEVICE_SERVER_COUNT_MASK;
				if ((count < 0) || (count > (int)sizeof(buf->data)))
				{
					trace(tl_error, "count invalid :%d, return 0 bytes read indicating fake connection closed", count);
					return 0;
				}
			}

			if(count - buf->len > 0)
			{
				ret = tp_read(fd, (unsigned char *)&buf->data[buf->len], count - buf->len);
				/* possibly we could get ret==0 buf connection is not closed ? */
				if(ret <= 0 )
				{
					return ret;
				}

				buf->len += ret;
				if(buf->len < count)
				{
					/* incomplete */
					*type = DEVICE_SERVER_NONE;
					trace(tl_data, "read_dserver(): Incomplete packet, count=%d, returning %d bytes read\n", count, ret);
					return ret;
				}
			}

			/* we have a complete packet */
			if(trace_level >= tl_data)
			{
				trace(tl_data, "read_dserver(): header=%04x, %d bytes", header, buf->len);
				dump_buf(trace_file, buf->data, buf->len);
			}

			if(header & DEVICE_SERVER_WINDOW)
			{
				out_window_size += 512;
			}


			if(header & DEVICE_SERVER_COMMAND)
			{
				trace(tl_data, "read_dserver(): Received dserver command with count=%d\n", count);
				if(count == 0)
				{
					/* was just window update */
					trace(tl_data, "read_dserver(): window update");
					*type = DEVICE_SERVER_NONE;
				}
				else
				{
					*type = buf->data [0];

					// if not an immediate command then update window size
					if ( (count > 0) && !(*type & DS_IMM_CMD_MASK))
					{
						trace(tl_data, "read_dserver(): In-line command - count = %d, cmd = %s", count, get_msg(*type, dserver_tcp_msg_list));
						update_input_window(count + 2);
						// if we were disconnected then return 0 bytes read indicating connection closed
						if (dserver_connected == FALSE)
						{
							return 0;
						}
					}

					switch(*type)
					{
						case DS_TCP_SEND_BREAK:
						{	
							write_tty_ctrl(ptyxm_fd, TP_PROTO_GOT_BREAK, NULL, 0);
							break;
						}
						case DS_TCP_CD1400_STATE:
						{
							new_state = buf->data [1];
							trace(tl_debug, "CD1400_STATE new modem state : %02x = %cDTR %cRTS %cRI %cDCD %cDSR %cCTS\n",
						       new_state,
						       new_state & CD_CTS ? '+' : '-',
						       new_state & CD_DSR ? '+' : '-',
						       new_state & CD_RI  ? '+' : '-',
						       new_state & CD_CD  ? '+' : '-',
						       new_state & CD_DTR ? '+' : '-',
						       new_state & CD_RTS ? '+' : '-');

							/* always report state */
							write_tty_ctrl(ptyxm_fd, TP_PROTO_DCD_UP, &new_state, 1);
							break;
						}
						case DS_TCP_MST:
						{
							// for historical reasons, the bit definition for this
							// command is not the same as for DS_TCP_CD1400_STATE,
							// but the information is the same
							temp_state=buf->data[1];
							
							// let's convert the bits to the DS_TCP_CD1400_STATE format
							// so we can call the write_tty_ctrl routine the same way
		
							new_state=0;
				
							if (temp_state & MST_CTS)
								new_state |= CD_CTS;
							if (temp_state & MST_DSR)
								new_state |= CD_DSR;
						  	if (temp_state & MST_RI)
								new_state |= CD_RI;
						  	if (temp_state & MST_DCD)
								new_state |= CD_CD;
						  	if (temp_state & MST_DTR)
								new_state |= CD_DTR;
						  	if (temp_state & MST_RTS)
								new_state |= CD_RTS;

						  	trace(tl_debug, "MST new modem state : %02x = %cCTS %cDSR %cRI %cDCD %cDTR %cRTS\n",
								new_state,
								new_state & CD_CTS ? '+' : '-',
								new_state & CD_DSR ? '+' : '-',
								new_state & CD_RI  ? '+' : '-',
								new_state & CD_CD  ? '+' : '-',
								new_state & CD_DTR ? '+' : '-',
								new_state & CD_RTS ? '+' : '-');
						  
							/* always report state */
							write_tty_ctrl(ptyxm_fd, TP_PROTO_DCD_UP, &new_state, 1);
							break;
						}
						case DS_TCP_LSR_NODATA:
						{
							new_state = buf->data [1];
							trace(tl_debug, "new line state : %02x = %cOE %cPE %cFE %cBI\n",
								new_state,
								new_state & LSRNODATA_OE ? '+' : '-',
								new_state & LSRNODATA_PE ? '+' : '-',
								new_state & LSRNODATA_FE  ? '+' : '-',
								new_state & LSRNODATA_BI  ? '+' : '-');

							/* always report state */
							write_tty_ctrl(ptyxm_fd, TP_PROTO_LSRUPDATE, &new_state, 1);
							break;
						}	
						case DS_TCP_SYNC_FLUSH_END:
						{
							write_tty_ctrl(ptyxm_fd, TP_PROTO_SYNC_FLUSH_END, NULL, 0);
							flushing_net_read = 0;
							// reset out ack count if DS supports it since we just flushed read data
							if (DServerFeatures & TPSFEAT_SYNCFLUSHENDRESET)
							{
								 in_window_used = 0;		// also DS doesn't require ack for DS_TCP_SYNC_FLUSH_END
							}
							// inform driver tp connection is now up if it was waiting on
							// a pending read flush
							if (!tp_connected && is_tp_conn_avail())
							{
								// update driver that tp connection is up
								write_tty_ctrl(ptyxm_fd, TP_PROTO_TP_CONNECTION, &tp_connected, 1);
							}
							trace( tl_debug, "sync_flush_end" );
							break;
						}
						case DS_TCP_PORT_NUMBER:
						{
							dserver_port_no = buf->data[1];
							trace(tl_debug, "device server port number is %d", dserver_port_no);
							DServerFeatures=ntohl(get_tps_feat(&buf->data[0], buf->len));
							trace( tl_debug, "server features 0x%x\n",DServerFeatures );
	 
							if (state == s_wait_port_number)
							{
								// This was put back in to turn on reporting if Device Server process was
								//	restarted without closing and opening the tty port.
								if ( (DServerFeatures & TPSFEAT_IOLANDS) == 0 )
								{
									 ret = write_dserver_ctrl(dserver_fd, DS_TCP_STTY_REPORT_ON);
								}
								else
								{
									ret = write_dserver_ctrl(dserver_fd, DS_TCP_STTY_XREPORT_ON);
								}
								// if DS supports no UDP then don't setup UDP socket
								// A bit kludgy but we will use the TPSFEAT_NO_UDP bit to determine if we want to send 
								// the DS_TCP_FEATURE_ACK cmd back to the DS.
								if ( (ret >= 0) && (DServerFeatures & TPSFEAT_NO_UDP) )
								{
									ret = write_dserver_ctrl(dserver_fd, DS_TCP_FEATURE_ACK);
								}
								if ( ret >= 0)
								{ 
									// if we are not useing UDP then we are now in data transfer state 
									if ( IsNoUDP() )
									{
										state = s_data_transfer;
									}
									else
									{
										setup_udp_socket();
										// only set state to data transfer state if we
										// get a UDP response
										ret = write_dserver_udp_ctrl_wait(DS_UDP_RESET );
										if (!ret)
										{
											state = s_data_transfer;
										}
									}
								}
	
								if ( ret < 0 )
								{
									close_connection();
								}
								else
								{
									// inform driver tp connection is now up if it was waiting on
									// a pending read flush
									if (!tp_connected && is_tp_conn_avail() )
									{
										// update driver that tp connection is up
										write_tty_ctrl(ptyxm_fd, TP_PROTO_TP_CONNECTION, &tp_connected, 1);
	 
									}
									tcgetattr(ptyxm_fd, &dserver_termios_modes);
									set_dserver_modes(TRUE);
								}
							}
							break;
						}
						case DS_TCP_DRAIN_ACK:
						{
							write_tty_ctrl(ptyxm_fd, TP_PROTO_DRAIN_ACK, NULL, 0);
							pending_drain = 0;
							break;
						}
 				 		case (DS_UDP_KEEP_ALIVE | DS_IMM_CMD_MASK):
						{
							trace(tl_debug, "read_dserver(): Immediate DS_UDP_KEEP_ALIVE command, count = %d", count);

							// acknowledge we received keep-alive response
							keepalive_sent = 0;
							break;
						}
						default:
						{
							trace(tl_debug, "read_dserver(): unexpected message");
					}
					} // end of switch
				}
				buf->len = 0;
				return (2+count);
			}
			else if (state == s_data_transfer)
			{
				if (count > 0)
				{
					update_input_window(count + 2);
				}
				if (dserver_connected == FALSE)
				{
					return 0;
				}
				*type = DEVICE_SERVER_DATA;
				trace(tl_data, "read_dserver(): DATA count=%d bytes", count);
				buf->data[buf->len+1] = '\0';
				return count;
			}
			else
			{
				trace(tl_data, "read_dserver(): not in data transfer state, and not a command, throu data away\n");
				*type = DEVICE_SERVER_NONE;
				return count;
			}
		}
	}
}



/****************************************************************************
 *	Read an DEVICE_SERVER UDP message
 ****************************************************************************/
/*
	Arguments	: udp_response : pointer to response structure
	Return values	: OK or FAILURE
*/
int read_dserver_udp_message(udp_response_t *udp_response)
{
	int     		ret;
	socklen_t      	addr_size;
	unsigned char   buf [16];

	if ( ipv6_not_running )
	{
		addr_size = sizeof (dserver_addr);
		ret = recvfrom(udp_sock, (char *) buf, sizeof (buf), 0,
		                (struct sockaddr *) &dserver_addr, &addr_size );
	}
	else
	{
		addr_size = sizeof (dserver_addr6);
		ret = recvfrom(udp_sock, (char *) buf, sizeof (buf), 0,
		                (struct sockaddr *) &dserver_addr6, &addr_size );
	}

	if (ret <= 0)
	{
		trace(tl_error, "receiving UDP message: %s", strerror(errno) );
		return FAILURE;
	}
	else
	{
		/* cmd, seq */
		udp_response->command = buf[0];
		udp_response->sequence_no = buf[1];

		if(trace_level >= tl_data)
		{
			trace(tl_data, "read UDP data: %d bytes", ret);
			dump_buf(trace_file, buf, ret);
		}
		trace(tl_data, "read_dserver_udp_message(): %s, seq = %02x",
		      get_msg(udp_response->command, dserver_udp_msg_list), udp_response->sequence_no);

	}

	return OK;
}
/****************************************************************************
 *	Set up a UDP socket for out of band communication with DEVICE_SERVER
 ****************************************************************************/
/*
	Arguments	: None
	Return values	: OK or FAILURE
 
 	dserver_addr must be set before calling(in wait_for_connect).
	On successful exit udp_sock and udp_addr are set.
*/
int setup_udp_socket(void)
{
	int on = 1;

	if ( udp_sock != -1 )
		return OK;
	
	// If IPv6 is not running then try IPv4.
	if ( ipv6_not_running )
	{
		udp_sock = socket( AF_INET, SOCK_DGRAM, 0 );

		if (udp_sock == -1)
		{
			trace(tl_error, "UDP socket create failed: %s\n", strerror(errno) );
			return FAILURE;
		}

		memset( &udp_addr, 0, sizeof(udp_addr) );

		udp_addr.sin_family = AF_INET;
		memset(&udp_addr.sin_addr, 0, sizeof (struct in_addr));
		udp_addr.sin_port = htons(tcp_port);

		setsockopt( udp_sock, SOL_SOCKET, SO_REUSEADDR, (const char *) &on, sizeof(on));

		if ( bind( udp_sock, (struct sockaddr *) & udp_addr, sizeof(udp_addr) ) < 0 )
		{
			trace(tl_error, "udp bind failed: %d\n", errno );
			return FAILURE;
		}

		memset( &udp_addr, 0, sizeof(udp_addr) );
		udp_addr.sin_family = AF_INET;
		udp_addr.sin_port = htons(DEVICE_SERVER_UDP_PORT);
		memcpy(&udp_addr.sin_addr, &dserver_addr.sin_addr, sizeof (struct in_addr));
	}
	else
	{
		udp_sock = socket( AF_INET6, SOCK_DGRAM, 0 );
		if (udp_sock == -1)
		{
			trace(tl_error, "UDP socket create failed(6): %s\n", strerror(errno) );
			return FAILURE;
		}

		memset( &udp_addr6, 0, sizeof(udp_addr6) );

		udp_addr6.sin6_family = AF_INET6;
		memset(&udp_addr6.sin6_addr, 0, sizeof (struct in6_addr));
		udp_addr6.sin6_port = htons(tcp_port);

		setsockopt( udp_sock, SOL_SOCKET, SO_REUSEADDR, (const char *) &on, sizeof(on));

		if ( bind( udp_sock, (struct sockaddr *) & udp_addr6, sizeof(udp_addr6) ) < 0 )
		{
			trace(tl_error, "udp bind failed(6): %d\n", errno );
			return FAILURE;
		}

		memset( &udp_addr6, 0, sizeof(udp_addr6) );
		udp_addr6.sin6_family = AF_INET6;
		udp_addr6.sin6_port = htons(DEVICE_SERVER_UDP_PORT);
		memcpy (&udp_addr6.sin6_addr, &dserver_addr6.sin6_addr, sizeof (struct in6_addr));
	}
	trace(tl_debug, "created udp socket");
	return OK;
}


/****************************************************************************
 *	Write control command to an DEVICE_SERVER connection(in-band TCP messages)
 ****************************************************************************/
/*
	Arguments	: fd   : descriptor of socket
			  cmd  : command
		Return values	: -1 if error(errno is set), 
					or 0/+Ve number of bytes(errno EAGAIN),
					otherwise no of bytes written
*/
int 	write_dserver_ctrl(int fd, int cmd)
{
	int ret= -1;
	dserver_cmd_t dserver_cmd;
	feature_ack_cmd_t feat_ack_cmd;


	if (mode == pm_lite)
	{
		{
			return(FAILURE);
		}
	}

	if (cmd == DS_TCP_FEATURE_ACK)
	{
		uint32_t TpFeatureSupport;

		feat_ack_cmd.count = htons(DEVICE_SERVER_COMMAND | (sizeof(feature_ack_cmd_t) - 2));
		feat_ack_cmd.cmd = cmd;
		dserver_cmd.command = cmd;
		TpFeatureSupport = DServerFeatures;
		if (use_legacy_UDP)
		{
			TpFeatureSupport &= (TPSFEAT_MASK & ~TPSFEAT_NO_UDP);
		}
		else
		{  
			TpFeatureSupport &= TPSFEAT_MASK;
		}

		feat_ack_cmd.tpfeatures = htonl(TpFeatureSupport);
		ret = write_buf_raw(fd, (unsigned char *)&feat_ack_cmd, sizeof(feature_ack_cmd_t));
		out_window_size -= sizeof(feature_ack_cmd_t);
		trace(tl_debug, "write_dserver_ctrl(): Sent DS_TCP_FEATURE_ACK Cmd=0x%X, TpFeatureSupport= 0x%04X\n", DS_TCP_FEATURE_ACK, TpFeatureSupport );
	}
	else
	{
		dserver_cmd.count = htons(DEVICE_SERVER_COMMAND | 1);
		dserver_cmd.command = cmd;
		// check if this command is to be put int the packet forwarding buffer and if so try
		// else just send command to dserver
		if ( !( (info.dscmd == cmd) && ChkPutDsCmdPktFwdBuf(&info, (char *)&dserver_cmd, 3) ) )
		{
			// if the command was to go in the pkt fwd buf then forward now before sending cmd
			if (info.dscmd)
				forward_frame(&info);
			ret = write_buf_raw(fd, (unsigned char *)&dserver_cmd, 3);
			out_window_size -= 3;
		}
	}
	trace(tl_debug, "write_dserver_ctrl(): Sending device server inline command %s", get_msg(dserver_cmd.command,dserver_tcp_msg_list) );
	return ret;
}


// This function will send any TCP Immediate commands, These are the old UDP
// OOB commands plus a few new ones.
// Note, these commands do not use up any of the window count
//	Return values	: -1 if error(errno is set), 
//					or 0/+Ve number of bytes(errno EAGAIN),
//					otherwise no of bytes written
int write_dserver_imm_ctrl(int fd, int cmd,  unsigned char data)
{
	int      			ret = -1;
	dserver_cmd_t		dserver_cmd;
	keepalive_cmd_t keepalive_cmd;
	unsigned char 	null_ch = 0;

	if (mode == pm_lite)
	{
		if ((cmd == DS_UDP_KEEP_ALIVE)
		)
		{
			null_ch = DS_URG_TCP_PACKETMODE_ACK;
			ret = send(dserver_fd,(const char *) &null_ch, 1, MSG_OOB);
			if( ret != 1 )
				return FAILURE;
			else
				return OK;
		}
		else
			/* no other controls supported */
			return FAILURE;
	}	

	// if currently using UDP protocol then return error
	if (!IsNoUDP())
	{
		trace(tl_debug, "write_dserver_imm_ctrl() called and device server does not support these commands" );
		return(FAILURE);
	}

	trace(tl_debug, "write_dserver_imm_ctrl(): Sending device server IMMediate command %s", get_msg( (cmd & ~DS_IMM_CMD_MASK), dserver_udp_msg_list) );
	
	// if keep-alive command and non-zero keep-alive value then send immediate keep-alive command
	if ( (cmd == DS_UDP_KEEP_ALIVE) && keepalive_time )
	{
		keepalive_cmd.count = htons(DEVICE_SERVER_COMMAND  | (sizeof(keepalive_cmd_t) - 2));
		keepalive_cmd.cmd = cmd | DS_IMM_CMD_MASK;		// set immediate command
	
		keepalive_cmd.keepalive_data.interval = htons(keepalive_time);
		ret = write_buf_raw( fd, (unsigned char *)&keepalive_cmd, sizeof(keepalive_cmd_t) );
	}
	// else the rest of the commands are all one byte
	else
	{
		dserver_cmd.count = htons(DEVICE_SERVER_COMMAND  | 1);
		dserver_cmd.command = cmd | DS_IMM_CMD_MASK;		// set immediate command
		ret = write_buf_raw(fd, (unsigned char *)&dserver_cmd, 3 );
	}

	return(ret);
}


/****************************************************************************
 *	Write UDP control command to a DEVICE_SERVER connection
 ****************************************************************************/
/*
	Arguments	: command  : UDP command
	Return values	: error code
*/
int 	write_dserver_udp_ctrl(int command)
{
	unsigned char 	null_ch = 0;
	unsigned char 	message [3];
	int 		ret;

	if (mode == pm_lite)
	{
		if ((command == DS_UDP_KEEP_ALIVE)
		)
		{
			/* synchronous write? */
			/* ret = write_buf_raw(dserver_fd, &null_ch, 1);
						We used a null byte in-band before - however this appears in
						data. If we send a UDP DS_URG_TCP_PACKETMODE_ACK,
						this is ignored by JS  */

			null_ch = DS_URG_TCP_PACKETMODE_ACK;
			ret = send(dserver_fd,(const char *) &null_ch, 1, MSG_OOB);
			if( ret != 1 )
				return FAILURE;
			else
				return OK;
		}
		else
			/* no other controls supported */
			return FAILURE;
	}
	else    /* full mode */
	{
		if(udp_sock == -1)
		{
			trace(tl_error, "Current udp_sock is invalid");
			return FAILURE;
		}

		if(command == DS_UDP_RESET)
		{
			udp_sequence_no = 0;
		}

		message[0] = dserver_port_no;
		message[1] = udp_sequence_no;
		message[2] = command;

		trace(tl_data, "sending UDP message %s, seq = %02x",
		      get_msg(command, dserver_udp_msg_list), udp_sequence_no);

		// The udp control channel is not encrypted, as SSL does not support
		//  udp connections.
		if( ipv6_not_running )
		{
			if( sendto( udp_sock,(const char *) message,
			            sizeof(message), 0,(const struct sockaddr *) &udp_addr,
			            sizeof(udp_addr) ) < 0 )
			{
				trace(tl_error, "UDP sendto failed: %d", errno );
				return FAILURE;
			}
		}
		else
		{
			if( sendto( udp_sock,(const char *) message,
			            sizeof(message), 0,(const struct sockaddr *) &udp_addr6,
			            sizeof(udp_addr6) ) < 0 )
			{
				trace(tl_error, "UDP sendto failed: %d\n", errno );
				return -1;
			}
		}

	}
	return OK;
}


/****************************************************************************
 *	Write UDP control command to an DEVICE_SERVER connection and wait for response
 ****************************************************************************/
/*
	Arguments	: command      : UDP command to send
	Return values	: error code
*/
int 	write_dserver_udp_ctrl_wait(int command)
{
	int  ret;
	fd_set		read_set;
	struct timeval	tim;
	unsigned int 	tries;

	if ( (mode == pm_full) && (udp_sock == -1) )
	{
		 // if currently not using UDP then send immediate command
		 if ( IsNoUDP() && (dserver_fd != -1) )
		 {
			  return(write_dserver_imm_ctrl(dserver_fd, command, 0));
		 }
		 else
			  return FAILURE;
	}
	
	for(tries=0; tries < max_udp_retries; tries++)
	{
		ret = write_dserver_udp_ctrl(command);
		if (mode == pm_lite)
		{
			return ret;
		}

		if (ret == FAILURE)
		{
			/* socket not open or sendto failed */
//			close_connection();
//			return FAILURE;
			goto write_udp_fail;
		}

		/* wait for reply */

		FD_ZERO( &read_set );
		FD_SET( udp_sock, &read_set );

		tim.tv_sec  = udp_ack_timeout / 1000000;
		tim.tv_usec = udp_ack_timeout % 1000000;

		ret = select(FD_SETSIZE, &read_set, NULL, NULL, &tim );

		if (ret == 0)
		{
			trace(tl_debug, "timed out waiting for UDP response(%d) - %x",
			      udp_sequence_no, command );
		}
		else if (ret == -1)
		{
			trace (tl_error, "error on select for UDP receive: %s", strerror (errno));
//			return FAILURE;
			goto write_udp_fail_noclose;
		}

		if((udp_sock != -1) && FD_ISSET( udp_sock, &read_set) )
		{
			ret = read_dserver_udp_message(&udp_response);

			/* process the response */

			switch (udp_response.command)
			{
				case DS_UDP_SYN:
				{
					 /* This is really very bad */
					/* *** abort connection */
//					close_connection();
//					return FAILURE;
					goto write_udp_fail;
					break;
				}
				case DS_UDP_ACK:
				{
					if (udp_response.sequence_no == udp_sequence_no)
					{
						trace(tl_debug, "UDP ACK sequence ok(%d) - %x", udp_sequence_no, command );
						udp_sequence_no++;
						return OK;
					}
					else
					{
						trace(tl_error, "error in UDP ACK sequence no(%d : %d) - %x ",
								udp_sequence_no, udp_response.sequence_no, command );
					}
					break;
				}
				case DS_UDP_WAK:
				{
					if (udp_response.sequence_no == udp_sequence_no)
					{
						trace (tl_debug, "UDP WAK sequence ok");
						udp_sequence_no++;
						return OK;
					}
					else
					{
						trace (tl_error, "error in UDP WAK sequence no");
					}
					break;
				}
				case DS_UDP_NAK:
				{
					 trace (tl_error, "received NAK");
					 /* fairly bad too */
//					close_connection();
//					return FAILURE;
					goto write_udp_fail;
					break;
				}
			default:
			{
				trace (tl_error, "unexpected UDP message: %d", udp_response.command);
			}
			} // end switch
		}
	}

write_udp_fail:
	close_connection();

write_udp_fail_noclose:
	return FAILURE;
}

//*********************************************************************
//  write_dserver_data - write application data to dserver
//****************************************************************************/
//
//	Arguments	: fd   : descriptor of socket
//			  buf  : pointer to buffer
//	Return values	: 0,+Ve number of bytes( errno maybe EAGAIN),
//					otherwise -1( errno set )
//
int write_dserver_data(int fd, buf_t *dstbuf, unsigned char *srcbuf, int srcbuf_len)
{
	int ret;
	
	unsigned short header;

	if (mode == pm_lite)
	{
		{
			memcpy(&dstbuf->data, srcbuf, srcbuf_len);
			dstbuf->len = srcbuf_len;
			ret = write_buf_raw(fd, &dstbuf->data[ dstbuf->offset ], dstbuf->len);
		}
	}
	else
	{
		// if using new pkt fwd logic then srcbuf is already packetized
		if (UseNewPktFwdLogic(&info))
		{
			memcpy(&dstbuf->data, srcbuf, srcbuf_len);
			dstbuf->len = srcbuf_len;
			out_window_size -= srcbuf_len;
			ret = write_buf_raw(fd, &dstbuf->data[ dstbuf->offset ], dstbuf->len);
		}
		else
		{
			 header = srcbuf_len;
			 // insert header into dstbuf
			 dstbuf->len += 2;
			 dstbuf->data[0] = (header >> 8) & 0xff;
			 dstbuf->data[1] = (header) & 0xff;
	 
			 // copy data into dstbuf
			 memcpy(&dstbuf->data[2], srcbuf, srcbuf_len);
			 dstbuf->len = srcbuf_len + 2;
	 
			 out_window_size -= dstbuf->len;
			 ret = write_buf_raw(fd, &dstbuf->data[dstbuf->offset], dstbuf->len);
		}
	}

	/* handle incomplete writes */
	if (ret > 0)
	{
		if (ret != dstbuf->len)
		{
			trace(tl_error, "incomplete write on device server, req %d written %d\n",dstbuf->len, ret);
		}
		dstbuf->len -= ret;
		dstbuf->offset += ret;
	}

	return ret;
}


/****************************************************************************
 *	Write control command to a TTY connection(COMredirect mode)
 ****************************************************************************/
/*
	Arguments: 
				fd   : descriptor of socket
				cmd  : command
				data : optional parameter data
				len  : length of parameter data
	Return values	: error code
*/
int write_tty_ctrl(int fd, int cmd, unsigned char *data, unsigned int len)
{
	unsigned int new_state;
	unsigned int old_state;
	int open_wait, temp_iparm;
	int ret;

	trace(tl_debug, "write_tty_ctrl(): cmd=0x%x\n", cmd);

	switch ( cmd )
	{
		case TP_PROTO_GOT_BREAK:
		{
			ret = ioctl(fd, TCSBRK, NULL);
			if (ret)
			{
				trace(tl_error, "Error in ioctl TCSBRK %d %d\n", ret, errno);
//				exit(-1);
				return FAILURE;
			}
			break;
		}
		case TP_PROTO_LSRUPDATE:
		{
			if (data)
			{
				new_state = *data;
				trace(tl_debug, "write_tty_ctrl TP_PROTO_LSRUPDATE : %02x = %cOE %cPE %cFE %cBI\n",
				   new_state,
				   new_state & LSRNODATA_OE ? '+' : '-',
				   new_state & LSRNODATA_PE ? '+' : '-',
				   new_state & LSRNODATA_FE  ? '+' : '-',
				   new_state & LSRNODATA_BI  ? '+' : '-');
				ret = ioctl(fd, PTX_IOCLSET, &new_state);	// set line status info
				if (ret)
				{
					trace(tl_error, "Error in ioctl %d %d\n", ret, errno);
//					exit(-1);
					return FAILURE;
				}
			}
			break;
		}
		case TP_PROTO_DCD_UP:
		case TP_PROTO_DCD_DOWN:
		{
			if (data)
			{
				new_state = *data;
				ret = ioctl(fd, TIOCMGET, &current_state);	// get modem status
				old_state = current_state;
				if (ret)
				{
					trace(tl_error, "Error in tiocmget: %d %d\n", ret, errno);
//					exit(-1);
					return FAILURE;
				}

				trace( tl_debug, "write_tty_ctrl: new=%08x, old=%08x\n", new_state, old_state );
				//		  if (new_state & CD_DSR) current_state |= TIOCM_RTS; else current_state &= ~TIOCM_RTS;
				if (new_state & CD_RI)
					current_state |= TIOCM_RI;
			  	else
					 current_state &= ~TIOCM_RI;

				if (new_state & CD_CD)
					 current_state |= TIOCM_CD;
				else
					 current_state &= ~TIOCM_CD;
            
				if (DServerFeatures & TPSFEAT_IOLANDS)
				{
					if (new_state & CD_DTR)
						current_state |= TIOCM_DTR;
				  else
						current_state &= ~TIOCM_DTR;
				
				  if (new_state & CD_RTS)
						current_state |= TIOCM_RTS;
				  else
						current_state &= ~TIOCM_RTS;
				
				  if (new_state & CD_DSR)
						current_state |= TIOCM_DSR;
				  else
						current_state &= ~TIOCM_DSR;
				
				  if (new_state & CD_CTS)
						current_state |= TIOCM_CTS;
				  else
						current_state &= ~TIOCM_CTS;
				}
				else
				{
					if (new_state & CD_DTR)
						current_state |= TIOCM_DSR;
				  else
						current_state &= ~TIOCM_DSR;
				
				  if (new_state & CD_RTS)
						current_state |= TIOCM_CTS;
				  else
						current_state &= ~TIOCM_CTS;
				}
				trace( tl_debug, "write_tty_ctrl: curr=%08x\n", current_state );

				ret = ioctl(fd, PTX_IOCMSET, &current_state);	// set modem status
				if (ret)
				{
					trace(tl_debug, "Error in ioctl-4: %d %d\n", ret, errno);
//					exit(-1);
					return FAILURE;
				}

				if (!(current_state & TIOCM_CD) && (old_state & TIOCM_CD) &&
			        !(dserver_termios_modes.c_cflag & CLOCAL))
				{
					trace( tl_info, "Lost CD - close tty port" );
					close_tty_no_rm();
				}
			}
			break;
		}
		case TP_PROTO_DRAIN_ACK:
		{
			ret = ioctl(fd, PTX_IOCDRAINEND, 0);	// drain complete
			if (ret)
			{
				trace(tl_error, "Error in ioctl-PTX_IOCDRAINEND: %d %d %d\n", __LINE__,ret, errno);
//				exit(-1);
				return FAILURE;
			}
		
			break;
		}
		case TP_PROTO_SET_OPEN_WAIT:
		{
			// data points to open wait time 
			if (data)
			{
				open_wait = *data;
				if (open_wait >= 0)
				{
					temp_iparm =  open_wait*1000;
				}
				else
				{
					temp_iparm = open_wait;
				}
				ret = ioctl(ptyxm_fd,PTX_IOCOWTSET,&temp_iparm);	// give ptyx, the open wait time in msec
				if (ret)
				{
					trace(tl_error, "Error in ioctl-PTX_IOCOWTSET: %d %d %d\n", __LINE__,ret, errno);
//					exit(-1);
					return FAILURE;
				}
			}
			break;
		}
		case TP_PROTO_TP_CONNECTION:
		{
			// data points to network connect state
			if (data)
			{
				ret = ioctl(ptyxm_fd,PTX_IOCNETSTAT, data);		// give ptyx, the connection status
				if (ret)
				{
					trace(tl_error, "Error in ioctl-PTX_IOCNETSTAT: %d %d %d\n", __LINE__,ret, errno);
//					exit(-1);
					return FAILURE;
				}
			}
			break;
		}
		default:
		{
			 break;
		}
	}
	return OK;
}

/****************************************************************************
 *	Update input window
 ****************************************************************************/
/*
	Arguments	: n	: amount used up
	Return values	: none
*/
void update_input_window(int n)
{
	unsigned short	window_ack;
	
	trace(tl_data, "update_input_window(): cur=%d, inc=%d", in_window_used, n);

	in_window_used += n;

	while ( in_window_used > 512)
	{
		window_ack = htons(DEVICE_SERVER_COMMAND | DEVICE_SERVER_WINDOW);  /* 0xC000 */

		/* *** check dserver_out_buf is empty ! */
		if (write_buf_raw(dserver_fd,(unsigned char *)&window_ack, 2) < 0)
		{
			trace(tl_error, "update_input_window(): Error sending window update cmd, closing connection\n");
			close_connection();
		}
		in_window_used -= 512;
	}
}

void drop_modem_signals(void)
{
    int ret;
    
    ret = ioctl(ptyxm_fd, TIOCMGET, &current_state);	// get modem status
    if (ret)
    {
        trace(tl_error, "%s(): TIOCMGET error %d %d\n", __FUNCTION__, ret, errno);
    }
    else
    {
        current_state &= (~(TIOCM_DSR|TIOCM_CD|TIOCM_CTS));
        ret = ioctl(ptyxm_fd, PTX_IOCMSET, &current_state);	// set modem status
        if (ret)
        {
            trace(tl_error, "%s(): PTX_IOCMSET error %d %d\n", __FUNCTION__, ret, errno);
        }
    }
}

//****************************************************************************
// Set terminal modes on DEVICE_SERVER using previously stored local modes
// ****************************************************************************/
//
//	Arguments	: set_all : if true set all modes
//	Return values	: none
//
// Baud Rate represented as actual values, strings and hardware specific
//

void set_dserver_modes(int set_all)
{
	struct dservermode    dserver_mode_cmd;
	struct termios  *modes;
	int     iflag = 0, oflag = 0, lflag = 0, cflag = 0;
	buf_t       buf;
	int     ret;

	/* set up a device server set_modes command using our current modes */

	if (mode == pm_lite)
		return;


	modes = &dserver_termios_modes;

	dserver_mode_cmd.count = htons(DEVICE_SERVER_COMMAND |(dservermodesize-2) );
	dserver_mode_cmd.cmd = DS_TCP_STTY_SET;

#if 0
	{
		 char buf[512];
		sprintf(buf, "Iflag: %s%s%s%s%s%s%s%s%s%s%s%s%s\n",
	         (modes->c_iflag & IGNBRK)?"IGNBRK ":"",
	         (modes->c_iflag & BRKINT)?"BRKINT ":"",
	         (modes->c_iflag & IGNPAR)?"IGNPAR ":"",
	         (modes->c_iflag & PARMRK)?"PARMRK ":"",
	         (modes->c_iflag & INPCK)?"INPCK ":"",
	         (modes->c_iflag & ISTRIP)?"ISTRIP ":"",
	         (modes->c_iflag & INLCR)?"INLCR ":"",
	         (modes->c_iflag & IGNCR)?"IGNCR ":"",
	         (modes->c_iflag & ICRNL)?"ICRNL ":"",
	         (modes->c_iflag & IUCLC)?"IUCLC ":"",
	         (modes->c_iflag & IXON)?"IXON ":"",
	         (modes->c_iflag & IXANY)?"IXANY ":"",
	         (modes->c_iflag & IXON)?"IXON ":"");
		trace(tl_debug, buf);
	
		sprintf(buf, "OFlag: %s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",
	         (modes->c_oflag & OPOST) ?"OPOST ":"",
	         (modes->c_oflag & OLCUC) ?"OLCUC ":"",
	         (modes->c_oflag & ONLCR) ?"ONLCR ":"",
	         (modes->c_oflag & OCRNL) ?"OCRNL ":"",
	         (modes->c_oflag & ONOCR) ?"ONOCR ":"",
	         (modes->c_oflag & OLCUC) ?"OLCUC ":"",
	         (modes->c_oflag & ONLRET) ?"ONLRET ":"",
	         (modes->c_oflag & OFILL) ? "OFILL ":"",
	         (modes->c_oflag & OFDEL) ? "OFDEL ":"",
	         (modes->c_oflag & NLDLY) ? "NLDLY ":"",
	         (modes->c_oflag & CR1) ? "CR1 ":"",
	         (modes->c_oflag & CR2) ? "CR2 ":"",
	         (modes->c_oflag & TAB1) ? "TAB1 ":"",
	         (modes->c_oflag & TAB2) ? "TAB2 ":"",
			         (modes->c_oflag & VTDLY) ? "VTDLY ":"",
	         (modes->c_oflag & FFDLY) ? "FFDLY ":"");
		trace(tl_debug, buf);

		sprintf(buf,"Cflag: %s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",
			(modes->c_cflag & CLOCAL) ? "CLOCAL ":"",
			(modes->c_cflag & CREAD) ? "CREAD ":"",
			(modes->c_cflag & CSTOPB) ? "CSTOPB ":"",
			(modes->c_cflag & HUPCL) ? "HUPCL ":"",
			(modes->c_cflag & PARENB) ? "PARENB ":"",
			(modes->c_cflag & PARODD) ? "PARODD ":"",
			(modes->c_cflag & MTS_TCSETA_SYNC) ? "MTS_TCSETA_SYNC ":"",
			((modes->c_cflag & CBAUD) == B50) ? "B50 ":"",
			((modes->c_cflag & CBAUD) == B75) ? "B75 ":"",
			((modes->c_cflag & CBAUD) == B110) ? "B110 ":"",
			((modes->c_cflag & CBAUD) == B134) ? "B134 ":"",
			((modes->c_cflag & CBAUD) == B150) ? "B150 ":"",
			((modes->c_cflag & CBAUD) == B200) ? "B200 ":"",
			((modes->c_cflag & CBAUD) == B300) ? "B300 ":"",
			((modes->c_cflag & CBAUD) == B600) ? "B600 ":"",
			((modes->c_cflag & CBAUD) == B1200) ? "B1200 ":"",
			((modes->c_cflag & CBAUD) == B1800) ? "B1800 ":"",
			((modes->c_cflag & CBAUD) == B2400) ? "B2400 ":"",
			((modes->c_cflag & CBAUD) == B4800) ? "B4800 ":"",
			((modes->c_cflag & CBAUD) == B9600) ? "B9600 ":"",
			((modes->c_cflag & CBAUD) == B19200) ? "B19200 ":"",
			((modes->c_cflag & CBAUD) == B38400) ? "B38400 ":"",
			((modes->c_cflag & CBAUD) == B57600) ? "B57600 ":"" ,
			((modes->c_cflag & CBAUD) == B115200) ? "B115200 ":"" ,
			((modes->c_cflag & CBAUD) == B230400) ? "B230400 ":"" ,
			((modes->c_cflag & CS8) == CS5) ? "CS5 ":"",
			((modes->c_cflag & CS8) == CS6) ? "CS6 ":"",
			((modes->c_cflag & CS8) == CS7) ? "CS7 ":"",
			((modes->c_cflag & CS8) == CS8) ? "CS8 ":"",
			(modes->c_cflag & CRTSCTS) ? "CRTSCTS ":"");
		trace(tl_debug, buf);

		sprintf(buf,"Lflag: %s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",
	        (modes->c_lflag & ISIG) ? "ISIG ":"",
	        (modes->c_lflag & ICANON) ? "ICANON ":"",
	        (modes->c_lflag & XCASE) ? "XCASE ":"",
	        (modes->c_lflag & ECHO) ? "ECHO ":"",
	        (modes->c_lflag & ECHOE) ? "ECHOE ":"",
	        (modes->c_lflag & ECHONL) ? "ECHONL ":"",
	        (modes->c_lflag & NOFLSH) ? "NOFLSH ":"",
	        (modes->c_lflag & ECHOCTL) ? "ECHOCTL ":"",
	        (modes->c_lflag & ECHOPRT) ? "ECHOPRT ":"",
	        (modes->c_lflag & ECHOKE) ? "ECHOKE ":"",
	        (modes->c_lflag & FLUSHO) ? "FLUSHO ":"",
	        (modes->c_lflag & PENDIN) ? "PENDIN ":"",
	        (modes->c_lflag & TOSTOP) ? "TOSTOP ":"",
	        (modes->c_lflag & IEXTEN) ? "IEXTEN ":"");
		trace(tl_debug, buf);
	}
#endif

	/*
	** The Termios structures used by the host dont always match
	** the MTS ones so translate ALL elements
	** DO NOT OVERWRITE the Termios structures though
	** otherwise the reads will be wrong
	*/
	iflag = ((modes->c_iflag & IGNBRK) ? MTS_IGNBRK : 0) |
	        ((modes->c_iflag & BRKINT) ? MTS_BRKINT : 0) |
	        ((modes->c_iflag & IGNPAR) ? MTS_IGNPAR : 0) |
	        ((modes->c_iflag & PARMRK) ? MTS_PARMRK : 0) |
	        ((modes->c_iflag & INPCK ) ? MTS_INPCK  : 0) |
	        ((modes->c_iflag & ISTRIP) ? MTS_ISTRIP : 0) |
	        ((modes->c_iflag & INLCR ) ? MTS_INLCR  : 0) |
	        ((modes->c_iflag & IGNCR ) ? MTS_IGNCR  : 0) |
	        ((modes->c_iflag & ICRNL ) ? MTS_ICRNL  : 0) |
	        ((modes->c_iflag & IUCLC ) ? MTS_IUCLC  : 0) |
	        ((modes->c_iflag & IXON  ) ? MTS_IXON   : 0) |
	        ((modes->c_iflag & IXANY ) ? MTS_IXANY  : 0) |
	        ((modes->c_iflag & IXOFF ) ? MTS_IXOFF  : 0) |
	        ((modes->c_iflag & IMAXBEL) ? MTS_IMAXBEL : 0);

	oflag = ((modes->c_oflag & OPOST) ? MTS_OPOST : 0) |
	        ((modes->c_oflag & OLCUC) ? MTS_OLCUC : 0) |
	        ((modes->c_oflag & ONLCR) ? MTS_ONLCR : 0) |
	        ((modes->c_oflag & OCRNL) ? MTS_OCRNL : 0) |
	        ((modes->c_oflag & ONOCR) ? MTS_ONOCR : 0) |
	        ((modes->c_oflag & OLCUC) ? MTS_OLCUC : 0) |
	        ((modes->c_oflag & ONLRET) ? MTS_ONLRET : 0) |
	        ((modes->c_oflag & OFILL) ? MTS_OFILL : 0) |
	        ((modes->c_oflag & OFDEL) ? MTS_OFDEL : 0) |
	        ((modes->c_oflag & NLDLY) ? MTS_NLDLY : 0) |
	        ((modes->c_oflag & CR1) ? MTS_CR1 : 0) |
	        ((modes->c_oflag & CR2) ? MTS_CR2 : 0) |
	        ((modes->c_oflag & TAB1) ? MTS_TAB1 : 0) |
	        ((modes->c_oflag & TAB2) ? MTS_TAB2 : 0) |
	        ((modes->c_oflag & VTDLY) ? MTS_VTDLY : 0) |
	        ((modes->c_oflag & FFDLY) ? MTS_FFDLY : 0);

	cflag = 0 |
	        /*        ((modes->c_cflag & CLOCAL) ? MTS_CLOCAL : 0) | */
	        ((modes->c_cflag & CREAD) ? MTS_CREAD : 0) |
	        ((modes->c_cflag & CSTOPB) ? MTS_CSTOPB : 0) |
	        ((modes->c_cflag & HUPCL) ? MTS_HUPCL : 0) |
	        ((modes->c_cflag & PARENB) ? MTS_PARENB : 0) |
	        ((modes->c_cflag & PARODD) ? MTS_PARODD : 0) |
	        /* ((modes->c_cflag & LOBLK) ? MTS_LOBLK : 0) | */
	        ((modes->c_cflag & MTS_TCSETA_SYNC) ? MTS_TCSETA_SYNC : 0) |  

	        (((modes->c_cflag & CBAUD) == B50) ? MTS_B50 : 0) |
	        (((modes->c_cflag & CBAUD) == B110) ? MTS_B110 : 0) |
	        (((modes->c_cflag & CBAUD) == B134) ? MTS_B134 : 0) |
	        (((modes->c_cflag & CBAUD) == B75)     ? MTS_B75     : 0) |
	        (((modes->c_cflag & CBAUD) == B150)    ? MTS_B150    : 0) |
	        (((modes->c_cflag & CBAUD) == B200)    ? MTS_B200    : 0) |
	        (((modes->c_cflag & CBAUD) == B300)    ? MTS_B300    : 0) |
	        (((modes->c_cflag & CBAUD) == B600)    ? MTS_B600    : 0) |
	        (((modes->c_cflag & CBAUD) == B1200)   ? MTS_B1200   : 0) |
	        (((modes->c_cflag & CBAUD) == B1800)   ? MTS_B1800   : 0) |
	        (((modes->c_cflag & CBAUD) == B2400)   ? MTS_B2400   : 0) |
	        (((modes->c_cflag & CBAUD) == B4800)   ? MTS_B4800   : 0) |
	        (((modes->c_cflag & CBAUD) == B9600)   ? MTS_B9600   : 0) |
	        (((modes->c_cflag & CBAUD) == B19200)  ? MTS_B19200  : 0) |
	        (((modes->c_cflag & CBAUD) == B38400)  ? MTS_B38400  : 0) |
	        (((modes->c_cflag & CBAUD) == B57600)  ? MTS_B57600  : 0) |
	        (((modes->c_cflag & CBAUD) == B115200) ? MTS_B115200 : 0) |
	        (((modes->c_cflag & CBAUD) == B230400) ? MTS_B230400 : 0) |

	        ((modes->c_cflag & CS6) ? MTS_CS6 : 0) |
	        ((modes->c_cflag & CS7) ? MTS_CS7 : 0) |
	        ((modes->c_cflag & CRTSCTS) ? (MTS_CTSFLOW | MTS_RTSFLOW) : 0);

	lflag = ((modes->c_lflag & ISIG) ? MTS_ISIG : 0) |
	        ((modes->c_lflag & ICANON) ? MTS_ICANON : 0) |
	        ((modes->c_lflag & XCASE) ? MTS_XCASE : 0) |
	        ((modes->c_lflag & ECHO) ? MTS_ECHO : 0) |
	        ((modes->c_lflag & ECHOE) ? MTS_ECHOE : 0) |
	        ((modes->c_lflag & ECHONL) ? MTS_ECHONL : 0) |
	        ((modes->c_lflag & NOFLSH) ? MTS_NOFLSH : 0) |
	        ((modes->c_lflag & ECHOCTL) ? MTS_ECHOCTL : 0) |
	        ((modes->c_lflag & ECHOPRT) ? MTS_ECHOPRT : 0) |
	        ((modes->c_lflag & ECHOKE) ? MTS_ECHOKE : 0) |
	        ((modes->c_lflag & FLUSHO) ? MTS_FLUSHO : 0) |
	        ((modes->c_lflag & PENDIN) ? MTS_PENDIN : 0) |
	        ((modes->c_lflag & TOSTOP) ? MTS_TOSTOP : 0) |
	        ((modes->c_lflag & IEXTEN) ? MTS_IEXTEN : 0);


	dserver_mode_cmd.iflag_low  = iflag & 0xFF;
	dserver_mode_cmd.iflag_high = (iflag >> 8) & 0xFF;

	dserver_mode_cmd.oflag_low  = oflag & 0xFF;
	dserver_mode_cmd.oflag_high = (oflag >> 8) & 0xFF;

	dserver_mode_cmd.cflag_low  = cflag & 0xFF;
	dserver_mode_cmd.cflag_high = (cflag >> 8) & 0xFF;

	dserver_mode_cmd.lflag_low  = lflag & 0xFF;
	dserver_mode_cmd.lflag_high = (lflag >> 8) & 0xFF;
	set_all = 0;
	if (set_all)
	{
		dserver_mode_cmd.imask_high = 0xFF;
		dserver_mode_cmd.imask_low  = 0xFF;
		dserver_mode_cmd.omask_high = 0xFF;
		dserver_mode_cmd.omask_low  = 0xFF;
		dserver_mode_cmd.cmask_high = 0xFF;
		dserver_mode_cmd.cmask_low  = 0xFF;
		dserver_mode_cmd.lmask_high = 0xFF;
		dserver_mode_cmd.lmask_low  = 0xFF;
	}
	else
	{
		dserver_mode_cmd.imask_low  = DEVICE_SERVER_IFLAGS & 0xFF;
		dserver_mode_cmd.imask_high = (DEVICE_SERVER_IFLAGS >> 8) & 0xFF;
		dserver_mode_cmd.omask_low  = DEVICE_SERVER_OFLAGS & 0xFF;
		dserver_mode_cmd.omask_high = (DEVICE_SERVER_OFLAGS >> 8) & 0xFF;
		dserver_mode_cmd.cmask_low  = DEVICE_SERVER_CFLAGS & 0xFF;
		dserver_mode_cmd.cmask_high = (DEVICE_SERVER_CFLAGS >> 8) & 0xFF;
		dserver_mode_cmd.lmask_low  = DEVICE_SERVER_LFLAGS & 0xFF;
		dserver_mode_cmd.lmask_high = (DEVICE_SERVER_LFLAGS >> 8) & 0xFF;
	}
	// if we are to put set tty command in buffer then attempt to do it
	// otherwise just send it to dserver
	if ( (info.dscmd != DS_TCP_STTY_SET) || !ChkPutDsCmdPktFwdBuf(&info, (char *)&dserver_mode_cmd, dservermodesize) )
	{
		forward_frame(&info);		// forward pkt fwd buf before sending command

		/* copy cmd to buffer */
		buf_init(&buf);
		buf_init(&dserver_in_buf);
	
		buf_append(&buf, (char *) &dserver_mode_cmd, dservermodesize);
	
		/* write to device server */
		out_window_size -= buf.len;
	
		if (tl_debug < trace_level)
		{
			my_hd(buf.data, buf.len);
		}
		
		trace(tl_debug, "set_dserver_modes() Sending DS_TCP_STTY_SET cmd");
		ret = write_buf_raw(dserver_fd, buf.data, buf.len);
	
		if (ret != buf.len)
		{
			trace(tl_error, "error writing tty modes, closing connection %d", ret);
			close_connection();
		}
		if ((DServerFeatures & TPSFEAT_IOLANDS)==0)
		{
			write_dserver_ctrl(dserver_fd, DS_TCP_STTY_REPORT_ON);
		}
	}
}

void set_RTS_DTR(int rts, int dtr)
{
	int ret;
	trace(tl_debug, "Update RTS/DTR (%x/%x):",rts,dtr);

	// are signals to be put into pkt fwd buf ?
	if (info.dscmd == (DS_TCP_SET_DTR + DS_TCP_CLR_DTR +
					   DS_TCP_SET_RTS + DS_TCP_CLR_RTS) )
	{
		info.dscmd = (rts ? DS_TCP_SET_RTS : DS_TCP_CLR_RTS);
		write_dserver_ctrl(dserver_fd, info.dscmd);		// write_dserver_ctrl will put command in pkt fwd buf
		info.dscmd = (dtr ? DS_TCP_SET_DTR : DS_TCP_CLR_DTR);
		write_dserver_ctrl(dserver_fd, info.dscmd);		// write_dserver_ctrl will put command in pkt fwd buf
	}
	else
	{
		ret = write_dserver_udp_ctrl_wait(rts ? DS_UDP_SET_RTS : DS_UDP_CLR_RTS);
		ret = write_dserver_udp_ctrl_wait(dtr ? DS_UDP_SET_DTR : DS_UDP_CLR_DTR);
		if ( ret < 0 )
		{
			trace(tl_error, "set_RTS_DTR(): Setting RTS/DTR signals failed, ret=%d, closing connection",ret);
			close_connection();
		}
	}
	trace(tl_debug, " %s_RTS", rts?"SET":"CLR");
	trace(tl_debug, " %s_DTR", dtr ? "SET" : "CLR");
}


//*****************************************************************************
//  tp_read -   reads data from network using SSL socket or tcp socket.
//
int  tp_read( int fd, unsigned char *buf, int size )
{
	int ret;

#ifdef  USE_SSL
	int err;

	if (fd == -1)
		return -1;

	if ( sslcfg.enabled )
	{
		ret = SSL_read( ssl, buf, size );
		switch( err = SSL_get_error( ssl, ret ) )
		{
			case SSL_ERROR_NONE:
			if ( ret <= 0 )
			{
				trace(tl_info, "SSL read - no data\n");
				return -1;
			}
			read_ssl = 0;
			break;

			case SSL_ERROR_WANT_READ:
			trace(tl_info, "SSL read - blocked\n");
			read_ssl = 1;
			break;

			default:
			trace(tl_error, "SSL read - error - %d; ret = %d\n", err, ret);
			if ( err == SSL_ERROR_SYSCALL && ret == 0 )
				return ret;
			else
				return -1;
		}
	}
	else
#endif  //USE_SSL

		ret = read( fd, buf, size );

	return(ret);
}


//*****************************************************************************
//  tp_read_wait -  reads data from fd until size bytes are received
//                  or ptim time-out occurs waiting to receive.  
//                  If NULL ptime given then it will wait forever
//
int tp_read_wait(int fd, unsigned char *buf, int size, struct timeval *ptim)
{
	struct timeval  tim_local, *ptim_local;
	int count, ret;
	int rcv_len;
	fd_set read_set;

	count = size;
	rcv_len = 0;

	while (1)
	{
		// read in  data
		ret = tp_read(fd, (unsigned char *)buf, count);
		if (ret <= 0 )
		{
			return(ret);
		}
		// finished receiving requested data
		if (ret == count)
		{
			rcv_len += ret;
			return(rcv_len);
		}
		count -= ret;
		rcv_len += ret;
		if (ptim)
		{
			tim_local.tv_sec = ptim->tv_sec;
			tim_local.tv_usec = ptim->tv_usec;
			ptim_local = &tim_local;
		}
		else
		{
			ptim_local = NULL;
		}
		// wait until header recieved 
		FD_ZERO( &read_set );
		FD_SET( fd, &read_set );
		ret = select(FD_SETSIZE, &read_set, NULL, NULL, ptim_local );

		// if time-out then return error
		if (ret == 0)
		{
			trace(tl_debug, "timed out waiting on select() for read data: %s", strerror(errno));
			return(-1);
		}
		else if (ret == -1)
		{
			trace(tl_debug, "error on select() data receive: %s", strerror(errno));
			return(-1);
		}
		if ( !(FD_ISSET( fd, &read_set)) )
		{
			return(-1);
		}
	} 	// while receiving data
}



//*****************************************************************************
// Returns < 0 if error
//
// host_port pointw to ipv4|hostname:portnum or [ipv6]:portnum
int get_tcp_hostport(char *host_port, char **TCPhost, char **TCPport)
{
   int port;
   char *phost = NULL;
   char *pport = NULL;

	if (host_port == NULL)
	{
		 return(-1);
	}
	// search for "[" to see if we have a literal IPv6 address
	phost = strchr((host_port), '[');
	// we have a literal IPv6 address
	if (phost != NULL)
	{
		// now search for "]", must have one
		pport = strchr(phost, ']');
		if ( pport == NULL)
		{
			 trace (tl_debug,"Missing ending bracket ] in IPv6 address\n");
			return(-1);		// error, missing ending bracket
		}
		else
		{
			// check if a port string exist
			if (strlen(pport) == 0)
			{
	 			 trace (tl_debug,"Missing port number \n");
				 return(-1);		// error, missing port number
			}
				
			pport = pport + 1;		// increment pass "]"
			// 
			if (*pport != ':')
			{
				 trace (tl_debug,"Missing port seperator : \n");
				return(-1);		// error, missing ending ":"
			}

			if (pport != NULL)
				 *pport = 0;				// null out ":"
			pport = pport + 1;				// increment pass ":"
			// check if a port string exist
			if (strlen(pport) == 0)
			{
				 trace (tl_debug,"Missing port number after : \n");
				return(-1);		// error, missing port number
			}
		}
		goto done_parse;
	}   // end of having IPv6 literal address
	
   // search for host and port seperator ":"
   pport = strchr(host_port, ':');
   // found ":", so first port is the host
   if (pport != NULL)
   {
		*pport = 0;				// null out ":"
		pport = pport + 1; 		// increment pointer to port
		phost = host_port;
		// if no port then set pointer back to NULL
		if (strlen(pport) == 0)
		{
			pport = NULL;
		}
		if (strlen(phost) == 0)
		{
		   trace (tl_debug,"Specified host can not be blank\n");
			return(-1);
		}
		if (strlen(phost) > MAX_HOST_NAMELEN)
		{
   		trace (tl_debug,"Host name too long, must be less than %d\n", MAX_HOST_NAMELEN);
			return(-1);
		}
	}
	// must be just the port, no host specified
	else
	{
		pport = host_port;
	}

done_parse:
	//  check port range
	if (pport != NULL)
	{
		port = atoi(pport);
		if ( (port < 0) || (port < WELLKNOWN) )
		{
			trace (tl_debug,"Port number must e in range: %d < TCP# > 65535\n", WELLKNOWN);
			return(-1);
		}
	}
	*TCPhost = phost;
	*TCPport = pport;
	return(0);
}

//*****************************************************************************
//  wait_for_tty_open_close 
//		- wait for the specified amount of time for the application to 
//		  open or close the port and take the appropriate action. 
//		  If a NULL ptime given then function will wait for ever for
//	  Returns -1 if tty close received or error
//			    0 if timeout
//			    1 if open recieved
//
int wait_for_tty_open_close(struct timeval *ptim)
{
	struct timeval  tim_local, *ptim_local;
	fd_set read_set;
	static char tty_data[1024];
	int len; 
	int ret=-1;

	if (ptim)
	{
		tim_local.tv_sec = ptim->tv_sec;
		tim_local.tv_usec = ptim->tv_usec;
		ptim_local = &tim_local;
	}
	else
	{
		ptim_local = NULL;
	}

	trace(tl_debug, "wait_for_tty_open_close(): Waiting for tty Open/Close ");

	while (tty_connected)
	{ 
		// wait for tty open/close or time out 
		FD_ZERO( &read_set );
		FD_SET( ptyxm_fd, &read_set );

		ret = select(FD_SETSIZE, &read_set, NULL, NULL, ptim );
		// if time-out then we will just return 0
		if (ret == 0)
		{
			trace(tl_debug, "wait_for_tty_open_close(): Timed-out on select() while waiting waiting for slave tty open,close ");
			break;
		}
		if (ret == -1)
		{
			trace(tl_debug, "wait_for_tty_open_close(): Error on select() while waiting for slave tty open,close: %s", strerror(errno));
			break;
		}
		if ( (ptyxm_fd != -1) && !(FD_ISSET( ptyxm_fd, &read_set)) )
		{
			ret = -1;
			break;
		}

		// read from master TTY
		len = read( ptyxm_fd, tty_data, sizeof(tty_data) );		
		/* If we get a read error, the most likely reason is that other
		   side has closed /dev/tx? . In any case, close the master
			side and let the main loop re-open it to prevent continually
			getting the error. 
		*/
		if (len == -1)
		{
			if (errno == EIO)
			{
				trace(tl_info, "wait_for_tty_open_close(): Port closed from other side - close our side");
				if (delay_close_tcp_flag)
				{
					delay_close_tcp_flag = FALSE;  
					trace(tl_info, "wait_for_tty_open_close(): closing TCP connection\n");
					close_connection();
					netfailed=0;
				}
			}
			else
			{
				trace(tl_error, "wait_for_tty_open_close(): Error reading from /dev/ptmx. errno: %d", errno);
			}
 			close_tty();
			ret = -1;
			break;
	  }
	  
	  if (!len)
	  {
		trace(tl_data, "wait_for_tty_open_close(): EOF in ptmx\n");
		exit(-1);
	  }

	  if (tty_data[0] & TIOCPKT_XCTRLSTATUS)
	  {
			typeof(ptyx_ctrl_status) temp_status;
			
			ret=ioctl(ptyxm_fd,PTX_IOCXCSGET,&temp_status);	// get extended control status
			if (ret)
			{
				trace(tl_error, "Error in ioctl-PTX_IOCXCSGET: %d %d %d\n", __LINE__,ret, errno);
				exit(-1);
			}
			ptyx_ctrl_status |= temp_status;
	  }
	  if (ptyx_ctrl_status & XCTRLSTATUS_CLOSE)
	  {
			trace(tl_info, "wait_for_tty_open_close(): XCTRLSTATUS_CLOSE recieved\n");
			ptyx_ctrl_status &= (~XCTRLSTATUS_CLOSE);
			slave_is_open = FALSE;
			// setup to close TCP connection on slave tty being closed
			//  as long as -nodisc not set
			if ( !nodisc_tcp && dserver_connected)
			{
				 delay_close_tcp_flag = TRUE;
			}
			ret = -1;
			break;
	  }
	  if (ptyx_ctrl_status & XCTRLSTATUS_OPEN)
	  {
			trace(tl_info, "wait_for_tty_open_close(): XCTRLSTATUS_OPEN recieved\n");
			ptyx_ctrl_status &= (~XCTRLSTATUS_OPEN);
			slave_is_open = TRUE;
			ret = 1;
			break;
	  }
	}  // end while tty_connected

	return(ret);
}

void handle_signal(int sig_num)
{
	// need to exit for these signals
	if ( (sig_num == SIGTERM) || (sig_num == SIGINT) )
	{
		trace( tl_error, "Exiting after catching signal SIGTERM or SIGINT ,signum=%d \n",  sig_num );
	}
	else
	{
		trace( tl_error, "Exiting after catching UNKNOWN signal, signum=%d \n", sig_num );	
	}
	close_tty();
	if (!ptyx_mode)
	{
		unlink(fixed_ttyname);
	}
	trace( tl_error, "Exiting after catching signal(%d)\n", sig_num );
	exit(-sig_num);		 
}


/*
	the format of the response containing the Server Features is:
	0x06
	PORT NUMBER		1 byte
	PRODUCT STR		variable length text, delimited with <space>
	<space>			1 byte
	VERSION STR		7 bytes
	SERVER FEAT		4 bytes (network order)
*/
uint32_t get_tps_feat(unsigned char *data,int datalen)
{
	int offset;
    uint32_t feat;
	
	offset=datalen-1;	// point to end of data
	while(offset >= 2 && data[offset]!=' ')
		offset--;
	offset+=8;	// point past the space and VERSION STR
	
	if (offset > datalen-4)
		return(0);	// SERVER FEAT doesn't exist in the response
	
	memcpy(&feat,&data[offset],sizeof(feat));
	return(feat);
}


int dserver_out_available(forwarding_info_t *pInfo)
{
	int val = 0;
	
	if (mode == pm_full)
	{
		val = out_window_size;	// # of bytes we can still send to the server
		if (pInfo->forwarding)
		{
			val -= pInfo->count;
		}
		if (val > (int)SMSIZEOF(buf_t2,data))
		{
			val = SMSIZEOF(buf_t2,data);
		}
		if (val < 0)
		{
			 val = 0;
		}
	}
	else
	{
		val=SMSIZEOF(buf_t2,data);
	}
	return(val);
}




#ifdef USE_SSL
//******************************************************************************
//
//  Used by SSL to copy the pass phrase into its own data space for use.

static int tp_pem_passwd_cb(char *buf, int size, int rwflag, void *password)
{
	UNUSED_VAR( rwflag );
	UNUSED_VAR( password );

	trace( tl_error, "Private Key Passphrase is not supported in COMredirect\n" );
	return( 0 );
}

//******************************************************************************

static RSA *tp_tmp_rsa_cb(SSL *s, int is_export, int keylength)
{
	static RSA *rsa_tmp=NULL;

	UNUSED_VAR( s );
	UNUSED_VAR( is_export );

	trace( tl_info,"generating tmp RSA key\n" );
	if ( rsa_tmp == NULL )
	{
		rsa_tmp=RSA_generate_key( keylength,RSA_F4,NULL,NULL );
	}
	return( rsa_tmp );
}
#endif  //USE_SSL


// verifies whether the COMredirect connection is available
// 
BOOL is_tp_conn_avail(void)
{
	BOOL bRet = 0;

	if (dserver_connected)		// dserver_connected will be set if TCP and SSL is up
	{
		bRet = TRUE;
	}

	// if TCP and/or SSL is up and were are in full mode, 
	// make sure we've finished detecting the protocol
	if (bRet  && mode == pm_full)
	{
		if (state != s_data_transfer)
		{
			bRet = FALSE;
		}
	}
	// set current COMredirect Connection connection state
	tp_connected = bRet;
	//debug, uncomment when required
//	trace(tl_debug, "is_tp_conn_avail(): COMredirect Connection is %s  ", (bRet ? "AVAILABLE" : "NOT AVAILABLE"));

	
	return bRet;
}


BOOL IsNoUDP(void)
{
	// check if dserver supports it and if we have it configured
	if ( (DServerFeatures & TPSFEAT_NO_UDP) && (!use_legacy_UDP) )
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

