/*******************************************************************************
 Module: ptyx.h
 
 Description: COMredirect common driver defines
 
 Copyright (c) Black Box Corporation 1999-2009
 All rights reserved
 
*******************************************************************************/

/* Set of debugging defines */
#define PTYX_DEBUG_OPEN					0x00000001
#define PTYX_DEBUG_CLOSE				0x00000002
#define PTYX_DEBUG_WRITE				0x00000004
#define PTYX_DEBUG_IOCTL				0x00000008
#define PTYX_DEBUG_WAIT_UNTIL_SENT	0x00000010
#define PTYX_DEBUG_BLOCK_TIL_READY	0x00000020
#define PTYX_DEBUG_TERMIOS				0x00000040
#define PTYX_DEBUG_MISC					0x00000080
#define PTYX_DEBUG_SPIN_LOCK			0x00000100

// uncomment line below and set ptyx_debug_level for debug tracing 
//#define PTYX_DEBUG

extern unsigned long  ptyx_debug_level; 

#define PTYX_PARANOIA_CHECK


#if defined(PTYX_DEBUG)
#define ptyx_print(flag, expr)				\
{														\
	if ( (flag & ptyx_debug_level) != 0)	\
		printk expr ;								\
}
#else
#define ptyx_print(flag, expr) {}
#endif


#define	PTYX_LOCK(slock, flags)					\
		do {						\
  		ptyx_print(PTYX_DEBUG_SPIN_LOCK, ("PTYX_LOCK: %s:%d slock = 0x%p \n", __FUNCTION__, __LINE__, slock)); \
		spin_lock_irqsave(slock, flags); \
		} while (0)
		
#define	PTYX_UNLOCK(slock, flags)					\
		do {						\
		spin_unlock_irqrestore(slock, flags); \
  		ptyx_print(PTYX_DEBUG_SPIN_LOCK, ("PTYX_UNLOCK: %s:%d slock = 0x%p \n", __FUNCTION__, __LINE__, slock)); \
		} while (0)





#ifndef USF_CLOSING_WAIT_NONE
#define USF_CLOSING_WAIT_NONE   (~0U)
#endif

#define msec_to_jiffies(x)  ((x) * HZ / 1000)

#define	DEFAULT_CLOSE_DELAY			500		// 0.5 sec
#define DEFAULT_CLOSING_WAIT_TIME	30000		// 30 sec


struct ptyx_struct 
{
	int	magic;
	wait_queue_head_t open_wait;
	unsigned int modem_status;
	struct serial_icounter_struct icount;
	wait_queue_head_t delta_msr_wait;
	int delta_msr_wait_done;
	int blocked_open; 				// # of blocked opens
	unsigned int close_delay;		// in msec.  If set and there are blocked opens, time to wait before signaling block processes to recheck for DCD 
	unsigned int closing_wait;	// in msec.  If set, time to wait for tx to drain before closing ; (~0=disabled, 0=forever) 
	unsigned int ctrl_status;		// ptyx master/slave driver control statuses
	int pending_drain_end;
	int network_status;				// network status (0=down, 1=up)
	int open_wait_time;				// if <0, don't wait for TCP connection.  If >=0, wait specified time for TCP connection
	int slave_active;				// indicates that at least 1 slave is active
	spinlock_t port_lock;			// lock to protect setting/reading of shared driver_data
};


// defines for ptyx_info flags
#define SLAVE_ACTIVE		0x00000001		// indicates that at least 1 slave is active
#define SLAVE_CLOSING	0x00000004		// slave tty is in the process of clossing





