#!/bin/sh
#
# Unistall script file for BLACK BOX COMredirect Linux driver
# 
# Copyright (c) 2003-2009  Black Box Corporation
#
# In principle, this should do the build-and-install. However, normally
# users are told to do the make, make install, etc.  steps separately
# This allows an operator to diagnose and solve problems in one of the
# steps more easily than if it's one "black box" of commands here. If
# you encounter problems, we recommend you follow these steps in this
# script one-by-one. 
#
###########################################################################################


###########################################################################################
#
#	This routine checks the details of any current installation of openssl and if the 
#	check fails, a variable will be sent to install a supplied copy of openssl
#

check_openssl()
{
	rm -f testopenssl
	make testopenssl
	result=$?
	if [ "$result" != "0" ]; then
		PROBLEM_TEXT=1
		NEED_SSL=yes
	fi
	
	if [ "$NEED_SSL" = "" ]; then
	  CHECK_LIBS=yes
	  export CHECK_LIBS
	  rm -f testopenssl
	  make testopenssl
	  result=$?
	  if [ "$result" != "0" ]; then
		PROBLEM_TEXT=2
		NEED_SSL=yes
	  else
		./testopenssl
		result=$?
		if [ "$result" = "2" ]; then
		  PROBLEM_TEXT=3
		  NEED_SSL=yes
		  
		else if [ "$result" = "3" ]; then
		  PROBLEM_TEXT=4
		  NEED_SSL=yes

		else if [ "$result" = "0" ]; then
			PROBLEM_TEXT=0
		else  
		  PROBLEM_TEXT=5
		  NEED_SSL=yes
		fi
		fi
		fi
	  fi
    fi

	return
}


###########################################################################################

# START of script.
KERNEL_VER_B=`uname -r | awk -F. '{printf "%d", $2}'`
KERNEL_VER_C=`uname -r | awk -F. '{printf "%d", $3}'`

if [ "$KERNEL_VER_B" -eq 6 ] && [ "$KERNEL_VER_C" -ge 28 ]; then
	 echo "ERROR: This version of COMredirect does not support Linux kernel"
	 echo "       versions greater than 2.6.28. COMredirect version 6.4.x"
	 echo "       or higher supports all Linux kernel versions 2.6.x or higher."
	 echo
	 exit 0
fi


DONE="not done"
while [ "$DONE" = "not done" ]; do
  echo 
  echo -n "Install SSL/TLS feature? (yes or no ): "
  read answer
  case "$answer" in
	yes|y)	SSL=1
			DONE=done
			;;

	no|n)	DONE="done"
			;;
		
	*)		;;	
  esac
done
if [ "$SSL" = "1" ]; then
	check_openssl	#check openssl function
fi

# determine architecture using uname -m
# need special config flag for application if on PPC64 or SPARC64
MACHINE="`uname -m`"

export SSL
export NEED_SSL
export PROBLEM_TEXT
export MACHINE

# If this is an rpm build, then set variables in files and exit to let rpm 
#  control the rest of the build process.

if [ "$1" = "rpm_build" ]; then
	echo $SSL > ./ssl
	echo $NEED_SSL > ./needssl
	echo $PROBLEM_TEXT > ./problem_text
	echo $MACHINE > ./machine
	exit 0
fi

###########################################################################################

make clean
make
if [ "$?" != "0" ]; then
	exit 2
fi
make install
depmod -a 

#remove testopenssl so that it will be rebuilt with the actual openssl being used.
if [ "$SSL" = "1" ]; then
  rm -f testopenssl
  CHECK_LIBS=yes
  export CHECK_LIBS
  make TESTOPENSSL=yes testopenssl
  ./testopenssl
  if [ "$?" != "0" ]; then
	  echo Openssl version still not current!!! 
	  echo "***********************************************"
	  exit 3
  fi
fi

./postinstall.sh

echo install complete

